#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2012 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" myhdl toVerilog conversion module.

"""
import sys
import math
import os
import textwrap

import inspect
import time
import ast
import string
from io import StringIO

from types import GeneratorType
import warnings

import myhdl
from myhdl import *
from myhdl import ToVerilogError, ToVerilogWarning
from myhdl._block import _Block
from myhdl._enum import EnumItemType
from myhdl._extractHierarchy import (_HierExtr, _isMem, _getMemInfo,
                                     _UserVerilogCode, _userCodeMap)
from myhdl._getHierarchy import _getHierarchy
from myhdl._instance import _Instantiator
from myhdl._Signal import _Signal, Constant
from myhdl._ShadowSignal import _TristateSignal, _TristateDriver
from myhdl.conversion._misc import (_error, _kind, _context,
                                    _ConversionMixin, _Label, _genUniqueSuffix, _isConstant)
from myhdl.conversion._analyze import (_analyzeSigs, _analyzeGens, _analyzeTopFunc,
                                       _Ram, _Rom)

_converting = 0
_profileFunc = None


def _checkArgs(arglist):
    for arg in arglist:
        if not isinstance(arg, (GeneratorType, _Instantiator, _UserVerilogCode)):
            raise ToVerilogError(_error.ArgType, arg)


def _flatten(*args):
    arglist = []
    for arg in args:
        if isinstance(arg, _Block):
            if arg.verilog_code is not None:
                arglist.append(arg.verilog_code)
                continue
            else:
                arg = arg.subs
        if id(arg) in _userCodeMap['verilog']:
            arglist.append(_userCodeMap['verilog'][id(arg)])
        elif isinstance(arg, (list, tuple, set)):
            for item in arg:
                arglist.extend(_flatten(item))
        else:
            arglist.append(arg)
    return arglist


def _makeDoc(doc, indent=''):
    if doc is None:
        return ''
    doc = inspect.cleandoc(doc)
    pre = '\n' + indent + '// '
    doc = '// ' + doc
    doc = doc.replace('\n', pre)
    return doc


class _ToVerilogConvertor(object):

    __slots__ = ("name",
                 "directory",
                 "timescale",
                 "standard",
                 "prefer_blocking_assignments",
                 "radix",
                 "header",
                 "no_myhdl_header",
                 "no_testbench",
                 "portmap",
                 "trace",
                 "initial_values"
                 )

    def __init__(self):
        self.name = None
        self.directory = None
        self.timescale = "1ns/10ps"
        self.standard = '2001'
        self.prefer_blocking_assignments = True
        self.radix = ''
        self.header = ''
        self.no_myhdl_header = False
        self.no_testbench = False
        self.trace = False
        self.initial_values = False

    def __call__(self, func, *args, **kwargs):
        global _converting
        if _converting:
            return func(*args, **kwargs)  # skip
        else:
            # clean start
            sys.setprofile(None)
        from myhdl import _traceSignals
        if _traceSignals._tracing:
            raise ToVerilogError("Cannot use toVerilog while tracing signals")
        if not isinstance(func, _Block):
            if not callable(func):
                raise ToVerilogError(_error.FirstArgType, "got %s" % type(func))

        _converting = 1
        if self.name is None:
            name = func.__name__
            if isinstance(func, _Block):
                name = func.func.__name__
        else:
            name = str(self.name)

        if isinstance(func, _Block):
            try:
                h = _getHierarchy(name, func)
            finally:
                _converting = 0
        else:
            warnings.warn(
                "\n    toVerilog(): Deprecated usage: See http://dev.myhdl.org/meps/mep-114.html",
                stacklevel=2,
                category=DeprecationWarning,
            )
            try:
                h = _HierExtr(name, func, *args, **kwargs)
            finally:
                _converting = 0

        if self.directory is None:
            directory = ''
        else:
            directory = self.directory

        vfilename = name + ".v"
        vpath = os.path.join(directory, vfilename)
        vfile = open(vpath, 'w')

        ### initialize properly ###
        _genUniqueSuffix.reset()

        arglist = _flatten(h.top)
        # print h.top
        _checkArgs(arglist)
        genlist = _analyzeGens(arglist, h.absnames)
        siglist, memlist = _analyzeSigs(h.hierarchy)
        _annotateTypes(genlist)

        # infer interface
        if isinstance(func, _Block):
            # infer interface after signals have been analyzed
            func._inferInterface()
            intf = func
        else:
            intf = _analyzeTopFunc(func, *args, **kwargs)

        intf.name = name

        doc = _makeDoc(inspect.getdoc(func))

        self._convert_filter(h, intf, siglist, memlist, genlist)

        _writeFileHeader(vfile, vpath, self.timescale)
        _writeModuleHeader(vfile, intf, doc)
        _writeSigDecls(vfile, intf, siglist, memlist)
        _convertGens(genlist, vfile)
        _writeModuleFooter(vfile)

        vfile.close()

        # don't write testbench if module has no ports
        if len(intf.argnames) > 0 and not toVerilog.no_testbench:
            tbpath = os.path.join(directory, "tb_" + vfilename)
            tbfile = open(tbpath, 'w')
            _writeTestBench(tbfile, intf, self.trace)
            tbfile.close()

        # build portmap for cosimulation
        portmap = {}
        for n, s in intf.argdict.items():
            if hasattr(s, 'driver'):
                portmap[n] = s.driver()
            else:
                portmap[n] = s
        self.portmap = portmap

        ### clean-up properly ###
        self._cleanup(siglist, memlist)

        return h.top

    def _cleanup(self, siglist, memlist):
        # clean up signals
        for sig in siglist:
            sig._clear()
        for mem in memlist:
            mem.name = None
            for s in mem.mem:
                s._clear()

        # clean up attributes
        self.name = None
        self.standard = '2001'
        self.prefer_blocking_assignments = True
        self.radix = ''
        self.header = ""
        self.no_myhdl_header = False
        self.no_testbench = False
        self.trace = False

    def _convert_filter(self, h, intf, siglist, memlist, genlist):
        # intended to be a entry point for other uses:
        #  code checking, optimizations, etc
        pass


toVerilog = _ToVerilogConvertor()

myhdl_header = """\
// File: $filename
// Generated by MyHDL $version
// Date: $date
"""


def _writeFileHeader(f, fn, ts):
    defs = dict(filename=fn,
                version=myhdl.__version__,
                date=f"   {time.asctime(time.gmtime())} UTC"
                )
    if not toVerilog.no_myhdl_header:
        print(string.Template(myhdl_header).substitute(defs), file=f)
    if toVerilog.header:
        print(string.Template(toVerilog.header).substitute(defs), file=f)
    print(file=f)
    print("`timescale %s" % ts, file=f)
    print(file=f)


def _writeModuleHeader(f, intf, doc):
    print("module %s (" % intf.name, file=f)
    b = StringIO()
    for portname in intf.argnames:
        print("    %s," % portname, file=b)
    print(b.getvalue()[:-2], file=f)
    b.close()
    print(");", file=f)
    print(doc, file=f)
    print(file=f)
    for portname in intf.argnames:
        s = intf.argdict[portname]
        if s._name is None:
            raise ToVerilogError(_error.ShadowingSignal, portname)
        if s._inList:
            raise ToVerilogError(_error.PortInList, portname)
        # make sure signal name is equal to its port name
        s._name = portname
        r = _getRangeString(s)
        p = _getSignString(s)
        if s._driven:
            if isinstance(s, _TristateSignal):
                print("inout %s%s%s;" % (p, r, portname), file=f)
            else:
                print("output %s%s%s;" % (p, r, portname), file=f)
            if s._driven == 'reg':
                print("reg %s%s%s;" % (p, r, portname), file=f)
            else:
                print("wire %s%s%s;" % (p, r, portname), file=f)
        else:
            if not s._read:
                warnings.warn("%s: %s" % (_error.UnusedPort, portname),
                              category=ToVerilogWarning
                              )
            print("input %s%s%s;" % (p, r, portname), file=f)
    print(file=f)


def _writeSigDecls(f, intf, siglist, memlist):
    constwires = []
    for s in siglist:
        if not s._used:
            continue

        if s._name in intf.argnames:
            continue

        if s._name.startswith('-- OpenPort'):
            # do not write a signal declaration
            continue

        r = _getRangeString(s)
        p = _getSignString(s)
        if s._driven:
            if not s._read and not isinstance(s, _TristateDriver):
                warnings.warn("%s: %s" % (_error.UnreadSignal, s._name),
                              category=ToVerilogWarning
                              )
            k = 'wire'
            if s._driven == 'reg':
                k = 'reg'
            # the following line implements initial value assignments
            # don't initial value "wire", inital assignment to a wire
            # equates to a continuous assignment [reference]
            if not toVerilog.initial_values or k == 'wire':
                print("%s %s%s%s;" % (k, p, r, s._name), file=f)
            else:
                if isinstance(s._init, myhdl._enum.EnumItemType):
                    print("%s %s%s%s = %s;" %
                          (k, p, r, s._name, s._init._toVerilog()), file=f)
                else:
                    print("%s %s%s%s = %s;" %
                          (k, p, r, s._name, _intRepr(s._init)), file=f)
        elif s._read:
            if isinstance(s, Constant):
                c = int(s.val)
                c_len = s._nrbits
                c_str = "%s" % c
                print("localparam %s%s = %s'd%s;" % (r, s._name, c_len, c_str), file=f)
            else:
                # the original exception
                # raise ToVerilogError(_error.UndrivenSignal, s._name)
                # changed to a warning and a continuous assignment to a wire
                warnings.warn("%s: %s" % (_error.UndrivenSignal, s._name),
                              category=ToVerilogWarning
                              )
                constwires.append(s)
                print("wire %s%s;" % (r, s._name), file=f)
    # print(file=f)
    for m in memlist:
        if not m._used:
            continue
        # infer attributes for the case of named signals in a list
        for s in m.mem:
            if not m._driven and s._driven:
                m._driven = s._driven
                # once suffices
                break
            if not m._read and s._read:
                m._read = s._read
                # once suffices
                break

        if not m._driven and not m._read:
            continue

        r = _getRangeString(m.elObj)
        p = _getSignString(m.elObj)
        k = 'wire'
        initial_assignments = None
        if m._driven:
            k = m._driven

            if toVerilog.initial_values and not k == 'wire':
                if all([each._init == m.mem[0]._init for each in m.mem]):
                    if toVerilog.initial_values == 'skip_zero_mem_init' and int(m.mem[0]._init) == 0:
                        pass
                    else:
                        initialize_block_name = ('INITIALIZE_' + m.name).upper()
                        _initial_assignments = (
                            '''
                            initial begin: %s
                                integer i;
                                for(i=0; i<%d; i=i+1) begin
                                    %s[i] = %s;
                                end
                            end
                            ''' % (initialize_block_name, len(m.mem), m.name,
                                   _intRepr(m.mem[0]._init)))

                        initial_assignments = (
                            textwrap.dedent(_initial_assignments))

                else:
                    val_assignments = '\n'.join(
                        ['    %s[%d] <= %s;' %
                         (m.name, n, _intRepr(each._init))
                         for n, each in enumerate(m.mem)])
                    initial_assignments = (
                        'initial begin\n' + val_assignments + '\nend')
            print("%s %s%s%s [0:%s-1];" % (k, p, r, m.name, m.depth), file=f)
        else:
            # remember for SystemVerilog, later
            # # can assume it is a localparam array
            # # build the initial values list
            # vals = []
            # w = m.mem[0]._nrbits
            # for s in m.mem:
            #     vals.append('{}\'d{}'.format(w, _intRepr(s._init)))
            #
            # print('localparam {} {} {} [0:{}-1] = \'{{{}}};'.format(p, r, m.name, m.depth, ', '.join(vals)), file=f)
            print('reg {}{} {} [0:{}-1];'.format(p, r, m.name, m.depth), file=f)
            val_assignments = '\n'.join(
                        ['    %s[%d] <= %s;' %
                         (m.name, n, _intRepr(each._init))
                         for n, each in enumerate(m.mem)])
            initial_assignments = (
                'initial begin\n' + val_assignments + '\nend')

        if initial_assignments is not None:
            print(initial_assignments, file=f)

    print(file=f)
    for s in constwires:
        if s._type in (bool, intbv):
            c = int(s.val)
        else:
            raise ToVerilogError("Unexpected type for constant signal", s._name)
        c_len = s._nrbits
        c_str = "%s" % c
        print("assign %s = %s'd%s;" % (s._name, c_len, c_str), file=f)
    # print(file=f)
    # shadow signal assignments
    for s in siglist:
        if hasattr(s, 'toVerilog') and s._driven:
            print(s.toVerilog(), file=f)
    print(file=f)


def _writeModuleFooter(f):
    print("endmodule", file=f)


def _writeTestBench(f, intf, trace=False):
    print("module tb_%s;" % intf.name, file=f)
    print(file=f)
    fr = StringIO()
    to = StringIO()
    pm = StringIO()
    for portname in intf.argnames:
        s = intf.argdict[portname]
        r = _getRangeString(s)
        if s._driven:
            print("wire %s%s;" % (r, portname), file=f)
            print("        %s," % portname, file=to)
        else:
            print("reg %s%s;" % (r, portname), file=f)
            print("        %s," % portname, file=fr)
        print("    %s," % portname, file=pm)
    print(file=f)
    print("initial begin", file=f)
    if trace:
        print('    $dumpfile("%s.vcd");' % intf.name, file=f)
        print('    $dumpvars(0, dut);', file=f)
    if fr.getvalue():
        print("    $from_myhdl(", file=f)
        print(fr.getvalue()[:-2], file=f)
        print("    );", file=f)
    if to.getvalue():
        print("    $to_myhdl(", file=f)
        print(to.getvalue()[:-2], file=f)
        print("    );", file=f)
    print("end", file=f)
    print(file=f)
    print("%s dut(" % intf.name, file=f)
    print(pm.getvalue()[:-2], file=f)
    print(");", file=f)
    print(file=f)
    print("endmodule", file=f)


def _getRangeString(s):
    if s._type is bool:
        return ''
    elif s._nrbits is not None:
        nrbits = s._nrbits
        return "[%s:0] " % (nrbits - 1)
    else:
        raise AssertionError


def _getSignString(s):
    if s._min is not None and s._min < 0:
        return "signed "
    else:
        return ''


def _intRepr(n, radix=''):
    # write size for large integers (beyond 32 bits signed)
    # with some safety margin
    # XXX signed indication 's' ???
    if isinstance(n, EnumItemType):
        return n._toVerilog()
    else:
        p = abs(n)
        size = ''
        num = str(p).rstrip('L')
        if radix == "hex" or p >= 2 ** 30:
            radix = "'h"
            num = hex(p)[2:].rstrip('L')
        if p >= 2 ** 30:
            size = int(math.ceil(math.log(p + 1, 2))) + 1  # sign bit!
    #            if not radix:
    #                radix = "'d"
        r = "%s%s%s" % (size, radix, num)
        if n < 0: # add brackets and sign on negative numbers
            r = "(-%s)" % r
        return r


def _convertGens(genlist, vfile):
    blockBuf = StringIO()
    funcBuf = StringIO()
    for tree in genlist:
        if isinstance(tree, _UserVerilogCode):
            blockBuf.write(str(tree))
            continue
        if tree.kind == _kind.ALWAYS:
            Visitor = _ConvertAlwaysVisitor
        elif tree.kind == _kind.INITIAL:
            Visitor = _ConvertInitialVisitor
        elif tree.kind == _kind.SIMPLE_ALWAYS_COMB:
            Visitor = _ConvertSimpleAlwaysCombVisitor
        elif tree.kind == _kind.ALWAYS_DECO:
            Visitor = _ConvertAlwaysDecoVisitor
        elif tree.kind == _kind.ALWAYS_SEQ:
            Visitor = _ConvertAlwaysSeqVisitor
        else: # ALWAYS_COMB
            Visitor = _ConvertAlwaysCombVisitor
        v = Visitor(tree, blockBuf, funcBuf)
        v.visit(tree)
    vfile.write(funcBuf.getvalue())
    funcBuf.close()
    vfile.write(blockBuf.getvalue())
    blockBuf.close()


opmap = {
    ast.Add: '+',
    ast.Sub: '-',
    ast.Mult: '*',
    ast.Div: '/',
    ast.Mod: '%',
    ast.Pow: '**',
    ast.LShift: '<<',
    ast.RShift: '>>>',
    ast.BitOr: '|',
    ast.BitAnd: '&',
    ast.BitXor: '^',
    ast.FloorDiv: '/',
    ast.Invert: '~',
    ast.Not: '!',
    ast.UAdd: '+',
    ast.USub: '-',
    ast.Eq: '==',
    ast.Gt: '>',
    ast.GtE: '>=',
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.NotEq: '!=',
    ast.And: '&&',
    ast.Or: '||',
}

nameconstant_map = {
    True: "1'b1",
    False: "1'b0",
    None: "'bz"
}


class _ConvertVisitor(ast.NodeVisitor, _ConversionMixin):

    def __init__(self, tree, buf):
        self.tree = tree
        self.buf = buf
        self.returnLabel = tree.name
        self.ind = ''
        self.isSigAss = False
        self.okSigAss = True
        self.labelStack = []
        self.context = _context.UNKNOWN

    def raiseError(self, node, kind, msg=""):
        lineno = self.getLineNo(node)
        info = "in file %s, line %s:\n    " % \
            (self.tree.sourcefile, self.tree.lineoffset + lineno)
        raise ToVerilogError(kind, msg, info)

    def write(self, arg):
        self.buf.write("%s" % arg)

    def writeline(self, nr=1):
        for i in range(nr):
            self.buf.write("\n%s" % self.ind)

    def writeDoc(self, node):
        assert hasattr(node, 'doc')
        doc = _makeDoc(node.doc, self.ind)
        self.write(doc)
        self.writeline()

    def indent(self):
        self.ind += ' ' * 4

    def dedent(self):
        self.ind = self.ind[:-4]

    def IntRepr(self, n, radix=''):
        return _intRepr(n, radix)

    def writeDeclaration(self, obj, name, dir):
        if dir:
            dir = dir + ' '
        if type(obj) is bool:
            self.write("%s%s" % (dir, name))
        elif isinstance(obj, int):
            if dir == "input ":
                self.write("input %s;" % name)
                self.writeline()
            self.write("integer %s" % name)
        elif isinstance(obj, _Ram):
            self.write("reg [%s-1:0] %s [0:%s-1]" % (obj.elObj._nrbits, name, obj.depth))
        elif hasattr(obj, '_nrbits'):
            s = ""
            if isinstance(obj, (intbv, _Signal)):
                if obj._min is not None and obj._min < 0:
                    s = "signed "
            self.write("%s%s[%s-1:0] %s" % (dir, s, obj._nrbits, name))
        else:
            raise AssertionError("var %s has unexpected type %s" % (name, type(obj)))
        # initialize regs
        # if dir == 'reg ' and not isinstance(obj, _Ram):
        # disable for cver
        if False:
            if isinstance(obj, EnumItemType):
                inival = obj._toVerilog()
            else:
                inival = int(obj)
            self.write(" = %s;" % inival)
        else:
            self.write(";")

    def writeDeclarations(self):
        for name, obj in self.tree.vardict.items():
            self.writeline()
            self.writeDeclaration(obj, name, "reg")

    def writeAlwaysHeader(self):
        assert self.tree.senslist
        senslist = self.tree.senslist
        self.write("always ")
        self.writeSensitivityList(senslist)
        self.write(" begin: %s" % self.tree.name)
        self.indent()

    def writeSensitivityList(self, senslist):
        sep = ', '
        if toVerilog.standard == '1995':
            sep = ' or '
        self.write("@(")
        for e in senslist[:-1]:
            self.write(e._toVerilog())
            self.write(sep)
        self.write(senslist[-1]._toVerilog())
        self.write(")")

    def visit_BinOp(self, node):
        if isinstance(node.op, ast.Mod) and self.context == _context.PRINT:
            self.visit(node.left)
            self.write(", ")
            self.visit(node.right)
        else:
            if isinstance(node.op, ast.RShift):
                # Additional cast to signed of the full expression
                # this is apparently required by cver - not sure if it
                # is actually required by standard Verilog.
                # It shouldn't hurt however.
                if node.signed:
                    self.write("$signed")
            self.context = None
            if node.signed:
                self.context = _context.SIGNED
            self.write("(")
            self.visit(node.left)
            self.write(" %s " % opmap[type(node.op)])
            self.visit(node.right)
            self.write(")")
            self.context = None

    def checkOpWithNegIntbv(self, node, op):
        if op in ("+", "-", "*", "~", "&&", "||", "!"):
            return
        if isinstance(node, ast.Name):
            o = node.obj
            if isinstance(o, (_Signal, intbv)) and o.min is not None and o.min < 0:
                self.raiseError(node, _error.NotSupported,
                                "negative intbv with operator %s" % op)

    def visit_BoolOp(self, node):
        self.write("(")
        self.visit(node.values[0])
        for n in node.values[1:]:
            self.write(" %s " % opmap[type(node.op)])
            self.visit(n)
        self.write(")")

    def visit_UnaryOp(self, node):
        self.write("(%s" % opmap[type(node.op)])
        self.visit(node.operand)
        self.write(")")

    def visit_Attribute(self, node):
        if isinstance(node.ctx, ast.Store):
            self.setAttr(node)
        else:
            self.getAttr(node)

    def setAttr(self, node):
        assert node.attr == 'next'
        self.isSigAss = self.okSigAss
        self.visit(node.value)

    def getAttr(self, node):
        if isinstance(node.value, ast.Subscript):
            self.setAttr(node)
            return

        assert isinstance(node.value, ast.Name), node.value
        n = node.value.id
        if n in self.tree.symdict:
            obj = self.tree.symdict[n]
        elif n in self.tree.vardict:
            obj = self.tree.vardict[n]
        else:
            raise AssertionError("object not found")
        if isinstance(obj, _Signal):
            if node.attr == 'next':
                self.isSigAss = self.okSigAss
                self.visit(node.value)
            elif node.attr in ('posedge', 'negedge'):
                self.write(node.attr)
                self.write(' ')
                self.visit(node.value)
            elif node.attr == 'val':
                self.visit(node.value)
        if isinstance(obj, (_Signal, intbv)):
            if node.attr in ('min', 'max'):
                self.write("%s" % node.obj)
        if isinstance(obj, EnumType):
            assert hasattr(obj, node.attr)
            e = getattr(obj, node.attr)
            self.write(e._toVerilog())

    def visit_Assert(self, node):
        self.write("if (")
        self.visit(node.test)
        self.write(" !== 1) begin")
        self.indent()
        self.writeline()
        self.write('$display("*** AssertionError ***");')
        # self.writeline()
        # self.write('$finish;')
        self.dedent()
        self.writeline()
        self.write("end")

    def visit_Assign(self, node):
        # shortcut for expansion of ROM in case statement
        if isinstance(node.value, ast.Subscript) and \
                      not isinstance(node.value.slice, ast.Slice) and \
                      isinstance(node.value.value.obj, _Rom):
            rom = node.value.value.obj.rom
#            self.write("// synthesis parallel_case full_case")
#            self.writeline()
            self.write("case (")
            self.visit(node.value.slice)
            self.write(")")
            self.indent()
            for i, n in enumerate(rom):
                self.writeline()
                if i == len(rom) - 1:
                    self.write("default: ")
                else:
                    self.write("%s: " % i)
                self.visit(node.targets[0])
                if self.isSigAss:
                    self.write(' <= ')
                    self.isSigAss = False
                else:
                    self.write(' = ')
                s = self.IntRepr(n)
                self.write("%s;" % s)
            self.dedent()
            self.writeline()
            self.write("endcase")
            return
        elif isinstance(node.value, ast.ListComp):
            # skip list comprehension assigns for now
            return
        # default behavior
        self.visit(node.targets[0])
        if self.isSigAss:
            self.write(' <= ')
            self.isSigAss = False
        else:
            self.write(' = ')
        self.visit(node.value)
        self.write(';')

    def visit_AugAssign(self, node, *args):
        # XXX apparently no signed context required for augmented assigns
        self.visit(node.target)
        self.write(" = ")
        self.visit(node.target)
        self.write(" %s " % opmap[type(node.op)])
        self.visit(node.value)
        self.write(";")

    def visit_Break(self, node,):
        self.write("disable %s;" % self.labelStack[-2])

    def visit_Call(self, node):
        self.context = None
        fn = node.func
        # assert isinstance(fn, astNode.Name)
        f = self.getObj(fn)

        if f is print:
            self.visit_Print(node)
            return

        opening, closing = '(', ')'
        if f is bool:
            self.write("(")
            self.visit(node.args[0])
            self.write(" != 0)")
            # self.write(" ? 1'b1 : 1'b0)")
            return
        elif f is len:
            val = self.getVal(node)
            self.require(node, val is not None, "cannot calculate len")
            self.write(repr(val))
            return
        elif f is now:
            self.write("$time")
            return
        elif f is ord:
            opening, closing = '', ''
            node.args[0].value = str(ord(node.args[0].value))
        elif f is int:
            opening, closing = '', ''
            # convert number argument to integer
            if isinstance(node.args[0], ast.Constant):
                node.args[0].value = int(node.args[0].value)
        elif f in (intbv, modbv):
            self.visit(node.args[0])
            return
        elif f == intbv.signed: # note equality comparison
            # comes from a getattr
            opening, closing = '', ''
            if not fn.value.signed:
                opening, closing = "$signed(", ")"
            self.write(opening)
            self.visit(fn.value)
            self.write(closing)
        elif (type(f) in (type,)) and issubclass(f, Exception):
            self.write(f.__name__)
        elif f in (posedge, negedge):
            opening, closing = ' ', ''
            self.write(f.__name__)
        elif f is concat:
            opening, closing = '{', '}'
        elif f is delay:
            self.visit(node.args[0])
            return
        elif hasattr(node, 'tree'):
            self.write(node.tree.name)
        else:
            self.write(f.__name__)
        if node.args:
            self.write(opening)
            self.visit(node.args[0])
            for arg in node.args[1:]:
                self.write(", ")
                self.visit(arg)
            self.write(closing)
        if hasattr(node, 'tree'):
            if node.tree.kind == _kind.TASK:
                Visitor = _ConvertTaskVisitor
            else:
                Visitor = _ConvertFunctionVisitor
            v = Visitor(node.tree, self.funcBuf)
            v.visit(node.tree)

    def visit_Compare(self, node):
        self.context = None
        if node.signed:
            self.context = _context.SIGNED
        self.write("(")
        self.visit(node.left)
        self.write(" %s " % opmap[type(node.ops[0])])
        self.visit(node.comparators[0])
        self.write(")")
        self.context = None

    def visit_Constant(self, node):
        if node.value is None:
            # NameConstant
            self.write(nameconstant_map[node.obj])
        elif isinstance(node.value, bool):
            self.write(nameconstant_map[node.obj])
        elif isinstance(node.value, int):
            # Num
            if self.context == _context.PRINT:
                self.write('"%s"' % node.value)
            else:
                self.write(self.IntRepr(node.value))
        elif isinstance(node.value, str):
            # Str
            s = node.value
            if self.context == _context.PRINT:
                self.write('"%s"' % s)
            elif len(s) == s.count('0') + s.count('1'):
                self.write("%s'b%s" % (len(s), s))
            else:
                self.write(s)

    def visit_Continue(self, node):
        self.write("disable %s;" % self.labelStack[-1])

    def visit_Expr(self, node):
        expr = node.value
        # docstrings on unofficial places
        if isinstance(expr, ast.Constant) and isinstance(expr.value, str):
            doc = _makeDoc(expr.value, self.ind)
            self.write(doc)
            return
        # skip extra semicolons
        if isinstance(expr, ast.Constant):
            return
        self.visit(expr)
        # ugly hack to detect an orphan "task" call
        if isinstance(expr, ast.Call) and hasattr(expr, 'tree'):
            self.write(';')

    def visit_IfExp(self, node):
        self.visit(node.test)
        self.write(' ? ')
        self.visit(node.body)
        self.write(' : ')
        self.visit(node.orelse)

    def visit_For(self, node):
        self.labelStack.append(node.breakLabel)
        self.labelStack.append(node.loopLabel)
        var = node.target.id
        cf = node.iter
        f = self.getObj(cf.func)
        args = cf.args
        assert len(args) <= 3
        if f is range:
            cmp = '<'
            op = '+'
            oneoff = ''
            if len(args) == 1:
                start, stop, step = None, args[0], None
            elif len(args) == 2:
                start, stop, step = args[0], args[1], None
            else:
                start, stop, step = args
        else: # downrange
            cmp = '>='
            op = '-'
            oneoff = '-1'
            if len(args) == 1:
                start, stop, step = args[0], None, None
            elif len(args) == 2:
                start, stop, step = args[0], args[1], None
            else:
                start, stop, step = args
        if node.breakLabel.isActive:
            self.write("begin: %s" % node.breakLabel)
            self.writeline()
        self.write("for (%s=" % var)
        if start is None:
            self.write("0")
        else:
            self.visit(start)
        self.write("%s; %s%s" % (oneoff, var, cmp))
        if stop is None:
            self.write("0")
        else:
            self.visit(stop)
        self.write("; %s=%s%s" % (var, var, op))
        if step is None:
            self.write("1")
        else:
            v = self.getVal(step)
            self.require(node, v >= 0, "step should be >= 0")
            self.visit(step)
        self.write(") begin")
        if node.loopLabel.isActive:
            self.write(": %s" % node.loopLabel)
        self.indent()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        if node.breakLabel.isActive:
            self.writeline()
            self.write("end")
        self.labelStack.pop()
        self.labelStack.pop()

    def visit_FunctionDef(self, node):
        raise AssertionError("To be implemented in subclass")

    def visit_If(self, node):
        if node.ignore:
            return
        if node.isCase:
            self.mapToCase(node)
        else:
            self.mapToIf(node)

    def visit_Match(self, node):
        self.write("case (")
        self.visit(node.subject)
        self.write(")")
        self.indent()
        for case in node.cases:
            self.visit(case)
            self.writeline()

        self.dedent()
        self.writeline()
        self.write("endcase")

    def visit_match_case(self, node):
        pattern = node.pattern
        self.visit(pattern)

        self.write(": begin ")
        self.indent()
        # Write all the multiple assignment per case
        for stmt in node.body:
            self.writeline()
            self.visit(stmt)
        self.dedent()
        self.writeline()
        self.write("end")

    def visit_MatchValue(self, node):
        item = node.value
        obj = self.getObj(item)

        if isinstance(obj, EnumItemType):
            itemRepr = obj._toVerilog()
        else:
            itemRepr = self.IntRepr(item.value, radix='hex')

        self.write(itemRepr)

    def visit_MatchSingleton(self, node):
        raise AssertionError("Unsupported Match type %s " % (type(node)))

    def visit_MatchSequence(self, node):
        raise AssertionError("Unsupported Match type %s " % (type(node)))

    def visit_MatchStar(self, node):
        raise AssertionError("Unsupported Match type %s " % (type(node)))

    def visit_MatchMapping(self, node):
        raise AssertionError("Unsupported Match type %s " % (type(node)))

    def visit_MatchClass(self, node):
        for pattern in node.patterns:
            self.visit(pattern)

    def visit_MatchAs(self, node):
        if node.name is None and  node.pattern is None:
            self.write("default")
        else:
            raise AssertionError("Unknown name %s or pattern %s" % (node.name, node.pattern))

    def visit_MatchOr(self, node):
        for i, pattern in enumerate(node.patterns):
            self.visit(pattern)
            if not i == len(node.patterns) - 1:
                self.write(" | ")

    def mapToCase(self, node, *args):
        var = node.caseVar
#        self.write("// synthesis parallel_case")
#        if node.isFullCase:
#            self.write(" full_case")
#        self.writeline()
        self.write("case (")
        self.visit(var)
        self.write(")")
        self.indent()
        for test, suite in node.tests:
            self.writeline()
            item = test.case[1]
            if isinstance(item, EnumItemType):
                self.write(item._toVerilog())
            else:
                self.write(self.IntRepr(item, radix='hex'))
            self.write(": begin")
            self.indent()
            self.visit_stmt(suite)
            self.dedent()
            self.writeline()
            self.write("end")
        if node.else_:
            self.writeline()
            self.write("default: begin")
            self.indent()
            self.visit_stmt(node.else_)
            self.dedent()
            self.writeline()
            self.write("end")
        self.dedent()
        self.writeline()
        self.write("endcase")

    def mapToIf(self, node, *args):
        first = True
        for test, suite in node.tests:
            if first:
                ifstring = "if ("
                first = False
            else:
                ifstring = "else if ("
                self.writeline()
            self.write(ifstring)
            self.visit(test)
            self.write(") begin")
            self.indent()
            self.visit_stmt(suite)
            self.dedent()
            self.writeline()
            self.write("end")
        if node.else_:
            self.writeline()
            self.write("else begin")
            self.indent()
            self.visit_stmt(node.else_)
            self.dedent()
            self.writeline()
            self.write("end")

    def visitKeyword(self, node, *args):
        self.visit(node.expr)

    def visit_Module(self, node, *args):
        for stmt in node.body:
            self.visit(stmt)

    def visit_ListComp(self, node):
        # do nothing
        pass

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.setName(node)
        else:
            self.getName(node)

    def setName(self, node):
        self.write(node.id)

    def getName(self, node):
        n = node.id
        addSignBit = False
        isMixedExpr = (not node.signed) and (self.context == _context.SIGNED)
        if n in self.tree.vardict:
            addSignBit = isMixedExpr
            s = n
        elif n in self.tree.argnames:
            assert n in self.tree.symdict
            addSignBit = isMixedExpr
            s = n
        elif n in self.tree.symdict:
            obj = self.tree.symdict[n]
            if isinstance(obj, bool):
                s = "1'b%s" % int(obj)
            elif isinstance(obj, int):
                s = self.IntRepr(obj)
            elif isinstance(obj, tuple): # Python3.9+ ast.Index replacement serves a tuple
                s = n
            elif isinstance(obj, _Signal):
                addSignBit = isMixedExpr
                s = str(obj)
            elif _isMem(obj):
                m = _getMemInfo(obj)
                assert m.name
                s = m.name
            elif isinstance(obj, EnumItemType):
                s = obj._toVerilog()
            elif (type(obj) in (type,)) and issubclass(obj, Exception):
                s = n
            else:
                self.raiseError(node, _error.UnsupportedType, "%s, %s %s" % (n, type(obj), obj))
        else:
            raise AssertionError("name ref: %s" % n)
        if addSignBit:
            self.write("$signed({1'b0, ")

        if s.startswith('--'):
            self.write(s.replace('--', '//'))
        else:
            self.write(s)
        if addSignBit:
            self.write("})")

    def visit_Pass(self, node):
        self.write("// pass")

    def visit_Print(self, node):
        argnr = 0
        for s in node.format:
            if isinstance(s, str):
                self.write('$write("%s");' % s)
            else:
                a = node.args[argnr]
                argnr += 1
                obj = a.obj
                if s.conv is int or isinstance(obj, int):
                    fs = "%0d"
                else:
                    fs = "%h"
                self.context = _context.PRINT
                if isinstance(obj, str):
                    self.write('$write(')
                    self.visit(a)
                    self.write(');')
                elif (s.conv is str) and isinstance(obj, bool):
                    self.write('if (')
                    self.visit(a)
                    self.write(')')
                    self.writeline()
                    self.write('    $write("True");')
                    self.writeline()
                    self.write('else')
                    self.writeline()
                    self.write('    $write("False");')
                elif isinstance(obj, EnumItemType):
                    tipe = obj._type
                    self.write('case (')
                    self.visit(a)
                    self.write(')')
                    self.indent()
                    for n in tipe._names:
                        self.writeline()
                        item = getattr(tipe, n)
                        self.write("'b%s: " % item._val)
                        self.write('$write("%s");' % n)
                    self.dedent()
                    self.writeline()
                    self.write("endcase")
                else:
                    self.write('$write("%s", ' % fs)
                    self.visit(a)
                    self.write(');')
                self.context = _context.UNKNOWN
            self.writeline()
        self.write('$write("\\n");')

    def visit_Raise(self, node):
        self.write("$finish;")

    def visit_Return(self, node):
        self.write("disable %s;" % self.returnLabel)

    def visit_Subscript(self, node):
        if isinstance(node.slice, ast.Slice):
            self.accessSlice(node)
        else:
            self.accessIndex(node)

    def accessSlice(self, node):
        if isinstance(node.value, ast.Call) and \
           node.value.func.obj in (intbv, modbv) and \
           _isConstant(node.value.args[0], self.tree.symdict):
            c = self.getVal(node)
            self.write("%s'h" % c._nrbits)
            self.write("%x" % c._val)
            return
        addSignBit = isinstance(node.ctx, ast.Load) and (self.context == _context.SIGNED)
        if addSignBit:
            self.write("$signed({1'b0, ")
        self.context = None
        self.visit(node.value)
        lower, upper = node.slice.lower, node.slice.upper
        # special shortcut case for [:] slice
        if lower is None and upper is None:
            return

        if isinstance(lower, ast.BinOp) and isinstance(lower.left, ast.Name) and isinstance(upper, ast.Name) and upper.id == lower.left.id and isinstance(lower.op, ast.Add):
            self.write("[")
            self.visit(upper)
            self.write("+:")
            self.visit(lower.right)
            self.write("]")
            return

        self.write("[")
        if lower is None:
            self.write("%s" % node.obj._nrbits)
        else:
            self.visit(lower)
        self.write("-1:")
        if upper is None:
            self.write("0")
        else:
            self.visit(upper)
        self.write("]")
        if addSignBit:
            self.write("})")

    def accessIndex(self, node):
        addSignBit = isinstance(node.ctx, ast.Load) and \
            (not node.signed) and \
            (self.context == _context.SIGNED)
        if addSignBit:
            self.write("$signed({1'b0, ")
        self.context = None
        self.visit(node.value)
        self.write("[")
        # assert len(node.subs) == 1
        if sys.version_info >= (3, 9, 0): # Python 3.9+: no ast.Index wrapper
            self.visit(node.slice)
        else:
            self.visit(node.slice.value)
        self.write("]")
        if addSignBit:
            self.write("})")

    def visit_stmt(self, body):
        for stmt in body:
            self.writeline()
            self.visit(stmt)
            # ugly hack to detect an orphan "task" call
            if isinstance(stmt, ast.Call) and hasattr(stmt, 'tree'):
                self.write(';')

    def visit_Tuple(self, node):
        assert self.context != None
        sep = ", "
        tpl = node.elts
        self.visit(tpl[0])
        for elt in tpl[1:]:
            self.write(sep)
            self.visit(elt)

    def visit_While(self, node):
        self.labelStack.append(node.breakLabel)
        self.labelStack.append(node.loopLabel)
        if node.breakLabel.isActive:
            self.write("begin: %s" % node.breakLabel)
            self.writeline()
        self.write("while (")
        self.visit(node.test)
        self.write(") begin")
        if node.loopLabel.isActive:
            self.write(": %s" % node.loopLabel)
        self.indent()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        if node.breakLabel.isActive:
            self.writeline()
            self.write("end")
        self.labelStack.pop()
        self.labelStack.pop()

    def visit_Yield(self, node):
        yieldObj = self.getObj(node.value)
        assert node.senslist
        senslist = node.senslist
        if isinstance(yieldObj, delay):
            self.write("# ")
            self.context = _context.YIELD
            self.visit(node.value)
            self.context = _context.UNKNOWN
            self.write(";")
        else:
            self.writeSensitivityList(senslist)
            self.write(";")


class _ConvertAlwaysVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        self.funcBuf = funcBuf

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        w = node.body[-1]
        y = w.body[0]
        if isinstance(y, ast.Expr):
            y = y.value
        assert isinstance(y, ast.Yield)
        self.writeAlwaysHeader()
        self.writeDeclarations()
        # assert isinstance(w.body, astNode.Stmt)
        for stmt in w.body[1:]:
            self.writeline()
            self.visit(stmt)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline(2)


class _ConvertInitialVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        self.funcBuf = funcBuf

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        self.write("initial begin: %s" % self.tree.name)
        self.indent()
        self.writeDeclarations()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline(2)


class _ConvertAlwaysCombVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        if toVerilog.prefer_blocking_assignments:
            self.okSigAss = False
        self.funcBuf = funcBuf

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        self.writeAlwaysHeader()
        self.writeDeclarations()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline(2)


class _ConvertSimpleAlwaysCombVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        self.funcBuf = funcBuf

    def visit_Attribute(self, node):
        if isinstance(node.ctx, ast.Store):
            # try intercepting '-- OpenPort' signals
            if isinstance(node.value, ast.Name):
                obj = self.tree.symdict[node.value.id]
                if obj._name.startswith('-- OpenPort'):
                    self.write('// ')

            self.write("assign ")
            self.visit(node.value)
        else:
            self.getAttr(node)

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        self.visit_stmt(node.body)
        self.writeline(2)


class _ConvertAlwaysDecoVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        self.funcBuf = funcBuf

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        self.writeAlwaysHeader()
        self.writeDeclarations()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline(2)


def _convertInitVal(reg, init):
    if isinstance(reg, _Signal):
        tipe = reg._type
    else:
        assert isinstance(reg, intbv)
        tipe = intbv
    if tipe is bool:
        v = '1' if init else '0'
    elif tipe is intbv:
        init = int(init)  # int representation
        v = "%s" % init if init is not None else "'bz"
    else:
        assert isinstance(init, EnumItemType)
        v = init._toVerilog()
    return v


class _ConvertAlwaysSeqVisitor(_ConvertVisitor):

    def __init__(self, tree, blockBuf, funcBuf):
        _ConvertVisitor.__init__(self, tree, blockBuf)
        self.funcBuf = funcBuf

    def visit_FunctionDef(self, node):
        self.writeDoc(node)
        self.writeAlwaysHeader()
        self.writeDeclarations()
        reset = self.tree.reset
        sigregs = self.tree.sigregs
        varregs = self.tree.varregs
        if reset is not None:
            self.writeline()
            self.write("if (%s == %s) begin" % (reset, int(reset.active)))
            self.indent()
            for s in sigregs:
                self.writeline()
                self.write("%s <= %s;" % (s, _convertInitVal(s, s._init)))
            for v in varregs:
                n, reg, init = v
                self.writeline()
                self.write("%s = %s;" % (n, _convertInitVal(reg, init)))
            self.dedent()
            self.writeline()
            self.write("end")
            self.writeline()
            self.write("else begin")
            self.indent()
        self.visit_stmt(node.body)
        self.dedent()
        if reset is not None:
            self.writeline()
            self.write("end")
            self.dedent()
        self.writeline()
        self.write("end")
        self.writeline(2)


class _ConvertFunctionVisitor(_ConvertVisitor):

    def __init__(self, tree, funcBuf):
        _ConvertVisitor.__init__(self, tree, funcBuf)
        self.returnObj = tree.returnObj
        self.returnLabel = _Label("RETURN")

    def writeOutputDeclaration(self):
        obj = self.tree.returnObj
        self.writeDeclaration(obj, self.tree.name, dir='')

    def writeInputDeclarations(self):
        for name in self.tree.argnames:
            obj = self.tree.symdict[name]
            self.writeline()
            self.writeDeclaration(obj, name, "input")

    def visit_FunctionDef(self, node):
        self.write("function ")
        self.writeOutputDeclaration()
        self.indent()
        self.writeInputDeclarations()
        self.writeDeclarations()
        self.dedent()
        self.writeline()
        self.write("begin: %s" % self.returnLabel)
        self.indent()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline()
        self.write("endfunction")
        self.writeline(2)

    def visit_Return(self, node):
        self.write("%s = " % self.tree.name)
        self.visit(node.value)
        self.write(";")
        self.writeline()
        self.write("disable %s;" % self.returnLabel)


class _ConvertTaskVisitor(_ConvertVisitor):

    def __init__(self, tree, funcBuf):
        _ConvertVisitor.__init__(self, tree, funcBuf)
        self.returnLabel = _Label("RETURN")

    def writeInterfaceDeclarations(self):
        for name in self.tree.argnames:
            obj = self.tree.symdict[name]
            output = name in self.tree.outputs
            input = name in self.tree.inputs
            inout = input and output
            dir = (inout and "inout") or (output and "output") or "input"
            self.writeline()
            self.writeDeclaration(obj, name, dir)

    def visit_FunctionDef(self, node):
        self.write("task %s;" % self.tree.name)
        self.indent()
        self.writeInterfaceDeclarations()
        self.writeDeclarations()
        self.dedent()
        self.writeline()
        self.write("begin: %s" % self.returnLabel)
        self.indent()
        self.visit_stmt(node.body)
        self.dedent()
        self.writeline()
        self.write("end")
        self.writeline()
        self.write("endtask")
        self.writeline(2)


def _maybeNegative(obj):
    if hasattr(obj, '_min') and (obj._min is not None) and (obj._min < 0):
        return True
    if isinstance(obj, int) and obj < 0:
        return True
    return False


class _AnnotateTypesVisitor(ast.NodeVisitor, _ConversionMixin):

    def __init__(self, tree):
        self.tree = tree

    def visit_FunctionDef(self, node):
        # don't visit arguments and decorators
        for stmt in node.body:
            self.visit(stmt)

    def visit_BinOp(self, node):
        self.visit(node.left)
        self.visit(node.right)
        node.signed = node.left.signed or node.right.signed
        # special treatement of subtraction unless in a top-level rhs
        if isinstance(node.op, ast.Sub) and not hasattr(node, 'isRhs'):
            node.signed = True

    def visit_BoolOp(self, node):
        for n in node.values:
            self.visit(n)
        node.signed = False

    def visit_UnaryOp(self, node):
        self.visit(node.operand)
        node.signed = node.operand.signed
        if isinstance(node.op, ast.USub):
            node.obj = int(-1)
            if isinstance(node.operand, ast.Constant):
                node.signed = True

    def visit_Attribute(self, node):
        if isinstance(node.ctx, ast.Store):
            self.setAttr(node)
        else:
            self.getAttr(node)

    def setAttr(self, node):
        self.visit(node.value)

    def getAttr(self, node):
        node.signed = False
        self.visit(node.value)

    def visit_Call(self, node):
        self.generic_visit(node)
        f = self.getObj(node.func)
        node.signed = False
        # suprize: identity comparison on unbound methods doesn't work in python 2.5??
        if f == intbv.signed:
            node.signed = True
        elif hasattr(node, 'tree'):
            v = _AnnotateTypesVisitor(node.tree)
            v.visit(node.tree)
            node.signed = _maybeNegative(node.tree.returnObj)

    def visit_Compare(self, node):
        node.signed = False
        # for n in ast.iter_child_nodes(node):
        for n in [node.left] + node.comparators:
            self.visit(n)
            if n.signed:
                node.signed = True

    def visit_If(self, node):
        if node.ignore:
            return
        self.generic_visit(node)

    def visit_Constant(self, node):
        node.signed = False

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            self.setName(node)
        else:
            self.getName(node)

    def setName(self, node):
        pass

    def getName(self, node):
        node.signed = _maybeNegative(node.obj)

    def visit_Subscript(self, node):
        if isinstance(node.slice, ast.Slice):
            self.accessSlice(node)
        else:
            self.accessIndex(node)

    def accessSlice(self, node):
        node.signed = False
        self.generic_visit(node)

    def accessIndex(self, node):
        node.signed = _maybeNegative(node.obj)
        self.generic_visit(node)

    def visit_Tuple(self, node):
        node.signed = False
        self.generic_visit(node)


def _annotateTypes(genlist):
    for tree in genlist:
        if isinstance(tree, _UserVerilogCode):
            continue
        v = _AnnotateTypesVisitor(tree)
        v.visit(tree)



from myhdl import (block, Signal, intbv, instance, delay)


@block
def dut():

    count = Signal(intbv(0, min=0, max=98))

    @instance
    def seq():
        count.next = 50
        for dummy in range(300):
            yield delay(10)
            print(count)
            if count - 1 < 0:
                count.next = 97
            else:
                count.next = count - 1

    return seq


def test_bug_aj1s():
    assert dut().verify_convert() == 0




from random import randrange

import myhdl
from myhdl import *

from bitonic import Array8Sorter, Array8Sorter_v

@block
def bench():

    n = 8
    w = 4

    a0, a1, a2, a3, a4, a5, a6, a7 = inputs = [Signal(intbv(0)[w:]) for i in range(n)]
    z0, z1, z2, z3, z4, z5, z6, z7 = outputs = [Signal(intbv(0)[w:]) for i in range(n)]


    inst = Array8Sorter_v(a0, a1, a2, a3, a4, a5, a6, a7,
                          z0, z1, z2, z3, z4, z5, z6, z7)

    @instance
    def check():
        for i in range(100):
            data = [randrange(2**w) for i in range(n)]
            for i in range(n):
                inputs[i].next = data[i]
            yield delay(10)
            data.sort()
            assert data == outputs

    return inst, check


def test_bench():
    bench().run_sim()

if __name__ == '__main__':
    test_bench()



import os
path = os.path
import unittest
from unittest import TestCase
import random
from random import randrange
random.seed(2)

from myhdl import (block, Signal, intbv, delay,
                   instance, StopSimulation, negedge)
from myhdl._Simulation import Simulation, Cosimulation

from .util import setupCosimulation

ACTIVE_LOW, INACTIVE_HIGH = 0, 1


@block
def inc_initial(count, enable, clock, reset, n):
    """ Incrementer with enable.
    
    count -- output
    enable -- control input, increment when 1
    clock -- clock input
    reset -- asynchronous reset input
    n -- counter max value
    """

    @instance
    def comb():
        for dummy in range(100):
            yield clock.posedge, reset.negedge
            if reset == ACTIVE_LOW:
                count.next = 0
            else:
                if enable:
                    count.next = (count + 1) % n
        raise StopSimulation

    return comb


objfile = "inc_initial_1.o"
analyze_cmd = "iverilog -o %s inc_initial_1.v tb_inc_initial_1.v" % objfile
vpifile = "../../../cosimulation/icarus/myhdl.vpi"
# vpifile = "cosimulation/icarus/myhdl.vpi"
simulate_cmd = "vvp -m %s %s" % (vpifile, objfile)


def top(name, count, enable, clock, reset, n, arch="myhdl"):
    if arch == "verilog":
        return setupCosimulation(**locals())
        if path.exists(objfile):
            os.remove(objfile)
        os.system(analyze_cmd)
        return Cosimulation(simulate_cmd, **locals())
    else:
        inc_initial_inst = inc_initial(count, enable, clock, reset, n)
        return inc_initial_inst


class TestInc_initial(TestCase):

    def clockGen(self, clock):
        while 1:
            yield delay(10)
            clock.next = not clock

    def stimulus(self, enable, clock, reset):
        reset.next = ACTIVE_LOW
        yield negedge(clock)
        reset.next = INACTIVE_HIGH
        for dummy in range(1000):
            enable.next = min(1, randrange(5))
            yield negedge(clock)
        raise StopSimulation

    def check(self, count, count_v, enable, clock, reset, n):
        expect = 0
        yield reset.posedge
        self.assertEqual(count, expect)
        self.assertEqual(count, count_v)
        while 1:
            yield clock.posedge
            if enable:
                expect = (expect + 1) % n
            yield delay(1)
            self.assertEqual(count, expect)
            self.assertEqual(count, count_v)

    def bench(self):

        m = 8
        n = 2 ** m

        count = Signal(intbv(0)[m:])
        count_v = Signal(intbv(0)[m:])
        enable, clock, reset = [Signal(bool()) for __ in range(3)]

        inc_initial_1 = top(top.__name__, count, enable, clock, reset, n=n).convert(hdl='Verilog')
        inc_initial_v = top(top.__name__, count_v, enable, clock, reset, n=n).convert(hdl='Verilog')
        clk_1 = self.clockGen(clock)
        st_1 = self.stimulus(enable, clock, reset)
        ch_1 = self.check(count, count_v, enable, clock, reset, n=n)

        sim = Simulation(inc_initial_1, inc_initial_v, clk_1, st_1, ch_1)
        return sim

    def test(self):
        """ Check increment operation """
        sim = self.bench()
        sim.run(quiet=1)


if __name__ == '__main__':
    unittest.main()




from myhdl import (block, Signal, intbv, always_comb, TristateSignal, toVHDL, toVerilog)

import pytest


@block
def issue_98(sda, scl, sda_i, sda_o, scl_i, scl_o):
    sda_d, scl_d = sda.driver(), scl.driver()

    @always_comb
    def hdl():
        sda_i.next = sda
        sda_d.next = 0 if not sda_o else None
        scl_i.next = scl
        scl_d.next = None if not scl_o else 1

    return hdl


@pytest.mark.xfail
def test_issue_98_1():
    sda_i, sda_o, scl_i, scl_o = [Signal(False) for __ in range(4)]
    sda, scl = [TristateSignal(False) for __ in range(2)]
    toVHDL.name = toVerilog.name = 'issue_98_1'
    assert issue_98(sda, scl, sda_i, sda_o, scl_i, scl_o).analyze_convert() == 0


@pytest.mark.xfail
def test_issue_98_2():
    sda_i, sda_o, scl_i, scl_o = [Signal(intbv(0)[2:0]) for __ in range(4)]
    sda, scl = [TristateSignal(intbv(0)[2:0]) for __ in range(2)]
    toVHDL.name = toVerilog.name = 'issue_98_2'
    assert issue_98(sda, scl, sda_i, sda_o, scl_i, scl_o).analyze_convert() == 0


@pytest.mark.xfail
def test_issue_98_3():
    sda_i, sda_o, scl_i, scl_o = [Signal(intbv(0)[1:0]) for __ in range(4)]
    sda, scl = [TristateSignal(intbv(0)[1:0]) for __ in range(2)]
    toVHDL.name = toVerilog.name = 'issue_98_3'
    assert issue_98(sda, scl, sda_i, sda_o, scl_i, scl_o).analyze_convert() == 0




#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.

#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" Run unit tests for Simulation """
import random
from random import randrange
from unittest import TestCase

from myhdl import (Signal, Simulation, SimulationError, StopSimulation, delay,
                   intbv, join, now)
from myhdl._Simulation import _error
from helpers import raises_kind

random.seed(1)  # random, but deterministic

QUIET = 1


class Shared:
    pass


class SimArgs(TestCase):
    """ Simulation arguments """

    def test1(self):
        with raises_kind(SimulationError, _error.ArgType):
            Simulation(None)

    def test2(self):

        def g():
            yield delay(10)

        i = g()
        with raises_kind(SimulationError, _error.DuplicatedArg):
            Simulation(i, i)


class YieldNone(TestCase):
    """ Basic test of yield None behavior """

    def test1(self):

        def stimulus():
            a = Signal(0)
            yield delay(10)
            a.next = 1
            yield None
            assert a.val == 0
            assert now() == 10
            yield delay(0)
            assert a.val == 1
            assert now() == 10

        Simulation(stimulus()).run(quiet=QUIET)

    def test2(self):

        def stimulus():
            a = Signal(0)
            yield delay(10)
            a.next = 1
            assert a.val == 0
            assert now() == 10
            yield None
            a.next = 0
            assert a.val == 0
            assert now() == 10
            yield None
            a.next = 1
            assert a.val == 0
            assert now() == 10
            yield delay(0)
            assert a.val == 1
            assert now() == 10

        Simulation(stimulus()).run(quiet=QUIET)

    def test3(self):

        def stimulus():
            a = Signal(0)
            yield delay(10)
            a.next = 1
            yield None, delay(10)
            assert a.val == 0
            assert now() == 10
            yield delay(0)
            assert a.val == 1
            assert now() == 10

        Simulation(stimulus()).run(quiet=QUIET)

    def test4(self):

        def stimulus():
            a = Signal(0)
            yield delay(10)

            def gen():
                yield delay(20)
                a.next = 1

            yield None, gen()
            assert a.val == 0
            assert now() == 10
            yield delay(25)
            assert a.val == 1

        Simulation(stimulus()).run(quiet=QUIET)


class JoinMix(TestCase):
    """ Test of joins mixed with other clauses """

    def test1(self):

        def stimulus():
            a = Signal(0)

            def gen():
                yield join(delay(10), delay(20))

            yield gen(), delay(5)
            assert now() == 5
            yield a
            raise AssertionError("Incorrect run")  # should not get here

        Simulation(stimulus()).run(quiet=QUIET)

    def test2(self):

        def stimulus():
            a = Signal(0)
            yield join(delay(10), delay(20)), delay(5)
            assert now() == 5
            yield a
            raise AssertionError("Incorrect run")  # should not get here

        Simulation(stimulus()).run(quiet=QUIET)

    def stimulus(self, a, b, c, d):
        yield delay(5)
        a.next = 1
        yield delay(5)
        a.next = 0
        b.next = 1
        yield delay(5)
        a.next = 1
        b.next = 0
        c.next = 1
        yield delay(5)
        a.next = 0
        b.next = 1
        c.next = 0
        d.next = 1

    def test3(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, b, c, d)
            assert now() == 20

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test4(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, b), join(c, d)
            assert now() == 10

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test5(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a), b, join(c, d)
            assert now() == 5

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test6(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, delay(20)), b, join(c, d)
            assert now() == 10

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test7(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, delay(30)), join(c, d)
            assert now() == 20

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test8(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, a.negedge)
            assert now() == 10

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test9(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, a.negedge, c.posedge)
            assert now() == 15

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test10(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, a)
            assert now() == 5

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)

    def test11(self):
        a, b, c, d = [Signal(0) for __ in range(4)]

        def response():
            yield join(a, b.posedge, b.negedge, a)
            assert now() == 15

        Simulation(self.stimulus(a, b, c, d), response()).run(quiet=QUIET)


class JoinedGen(TestCase):

    """ Basic test of yielding joined concurrent generators """

    def bench(self):
        sig1 = Signal(0)
        sig2 = Signal(0)
        td = 10

        def gen(s, n):
            for dummy in range(n - 1):
                yield delay(td)
            s.next = 1
            yield delay(td)

        for __ in range(10):
            offset = now()
            n0 = randrange(1, 50)
            n1 = randrange(1, 50)
            n2 = randrange(1, 50)
            sig1.next = 0
            sig2.next = 0
            yield join(delay(n0 * td), gen(sig1, n1), gen(sig2, n2))
            assert sig1.val == 1
            assert sig2.val == 1
            assert now() == offset + td * max(n0, n1, n2)

        raise StopSimulation("Joined concurrent generator yield")

    def testYieldJoinedGen(self):
        Simulation(self.bench()).run(quiet=QUIET)


class SignalUpdateFirst(TestCase):

    """ Check that signal updates are done first, as in VHDL """

    def bench(self):

        Q = Signal(0, delay=9)
        R = Signal(0, delay=10)
        S = Signal(0, delay=11)

        def process():
            Q.next = 0
            R.next = 0
            S.next = 0
            yield delay(50)
            Q.next = 1
            R.next = 1
            S.next = 1
            yield delay(10)
            assert Q.val == 1  # control
            assert R.val == 1  # actual check
            assert S.val == 0  # control
            yield delay(1)
            assert Q.val == 1  # control
            assert R.val == 1  # control
            assert S.val == 1  # control
            raise StopSimulation("Signal update test")

        return process()

    def testSignalUpdateFirst(self):
        Simulation(self.bench()).run(quiet=QUIET)


class YieldZeroDelay(TestCase):

    """ Basic test of yielding a zero delay """

    def bench(self):
        sig1 = Signal(0)
        sig2 = Signal(0)
        td = 10

        def gen(s, n):
            s.next = 0
            for dummy in range(n):
                yield delay(td)
            s.next = 1

        for __ in range(100):
            offset = now()
            n1 = randrange(2, 10)
            n2 = randrange(n1 + 1, 20)  # n2 > n1
            yield delay(0), gen(sig1, n1), gen(sig2, n2)
            assert sig1.val == 0
            assert sig2.val == 0
            assert now() == offset + 0
            yield sig1.posedge
            assert sig2.val == 0
            assert now() == offset + n1 * td
            yield sig2.posedge
            assert now() == offset + n2 * td

        raise StopSimulation("Zero delay yield")

    def testYieldZeroDelay(self):
        Simulation(self.bench()).run(quiet=QUIET)


class YieldConcurrentGen(TestCase):

    """ Basic test of yielding concurrent generators """

    def bench(self):
        sig1 = Signal(0)
        sig2 = Signal(0)
        td = 10

        def gen(s, n):
            s.next = 0
            for dummy in range(n):
                yield delay(td)
            s.next = 1

        for __ in range(100):
            offset = now()
            n1 = randrange(2, 10)
            n2 = randrange(n1 + 1, 20)  # n2 > n1
            yield delay(td), gen(sig1, n1), gen(sig2, n2)
            assert sig1.val == 0
            assert sig2.val == 0
            assert now() == offset + td
            yield sig1.posedge
            assert sig2.val == 0
            assert now() == offset + n1 * td
            yield sig2.posedge
            assert now() == offset + n2 * td

        raise StopSimulation("Concurrent generator yield")

    def testYieldConcurrentGen(self):
        Simulation(self.bench()).run(quiet=QUIET)


class YieldGen(TestCase):

    """ Basic test of yielding generators """

    def bench(self):

        clk = Signal(0)
        shared = Shared()
        shared.cnt = 0
        shared.i = 0
        expected = []
        nlists = []
        expectedCnt = 0
        for __ in range(300):
            l = []
            for __ in range(randrange(1, 6)):
                e = randrange(0, 5)
                l.append(e)
                expectedCnt += e
                expected.append(expectedCnt)
            nlists.append(l)

        def clkGen():
            while 1:
                yield delay(10)
                clk.next = 1
                yield delay(10)
                clk.next = 0

        def task(nlist):
            n = nlist.pop(0)
            for dummy in range(n):
                yield clk.posedge
                shared.cnt += 1
            assert shared.cnt == expected[shared.i]
            shared.i += 1
            if nlist:
                yield task(nlist)

        def module():
            for nlist in nlists:
                yield task(nlist)
            assert shared.cnt == expected[-1]
            raise StopSimulation("Generator yield")

        return(module(), clkGen())

    def testYieldGen(self):
        Simulation(self.bench()).run(quiet=QUIET)


class DeltaCycleOrder(TestCase):

    """ Check that delta cycle order does not matter """

    def bench(self, function):

        clk = Signal(0)
        a = Signal(0)
        b = Signal(0)
        c = Signal(0)
        d = Signal(0)
        z = Signal(0)
        delta = [Signal(0) for dummy in range(4)]
        inputs = Signal(intbv(0))
        s = [a, b, c, d]
        vectors = [intbv(j) for dummy in range(8) for j in range(16)]
        random.shuffle(vectors)
        index = list(range(4))

        def clkGen():
            while 1:
                yield delay(10)
                clk.next ^= 1

        def deltaGen():
            while 1:
                yield clk
                delta[0].next = clk.val
                yield delta[0]
                for i in range(1, 4):
                    delta[i].next = delta[i - 1].val
                    yield delta[i]

        def inGen(i):
            while 1:
                yield delta[i].posedge
                s[index[i]].next = inputs.val[index[i]]

        def logic():
            while 1:
                # yield a, b, c, d
                z.next = function(a.val, b.val, c.val, d.val)
                yield a, b, c, d

        def stimulus():
            for v in vectors:
                inputs.next = v
                random.shuffle(index)
                yield clk.posedge
                yield clk.posedge
                assert z.val == function(v[0], v[1], v[2], v[3])
            raise StopSimulation("Delta cycle order")

        inputGen = [inGen(i) for i in range(4)]
        instance = [clkGen(), deltaGen(), logic(), stimulus(), inputGen]
        return instance

    def testAnd(self):

        def andFunction(a, b, c, d):
            return a & b & c & d

        Simulation(self.bench(andFunction)).run(quiet=QUIET)

    def testOr(self):

        def orFunction(a, b, c, d):
            return a | b | c | d

        Simulation(self.bench(orFunction)).run(quiet=QUIET)

    def testXor(self):

        def xorFunction(a, b, c, d):
            return a ^ b ^ c ^ d

        Simulation(self.bench(xorFunction)).run(quiet=QUIET)

    def testMux(self):

        def muxFunction(a, b, c, d):
            if c:
                return a
            else:
                return b

        Simulation(self.bench(muxFunction)).run(quiet=QUIET)

    def testLogic(self):

        def function(a, b, c, d):
            return not (a & (not b)) | ((not c) & d)

        Simulation(self.bench(function)).run(quiet=QUIET)


class DeltaCycleRace(TestCase):

    """ Check that delta cycle races are like in VHDL """

    def bench(self):

        uprange = range(300)
        msig = Signal(uprange[0])
        ssig = [Signal(uprange[-1]) for __ in range(2)]
        dsig = [Signal(uprange[0]) for __ in range(2)]
        clk = Signal(0)
        deltaClk = Signal(0)
        shared = Shared()
        shared.t = now()

        def clkGen():
            for dummy in uprange[:-1]:
                yield delay(10)
                clk.next = 1
                yield delay(10)
                clk.next = 0

        def deltaClkGen():
            while 1:
                yield clk
                deltaClk.next = clk.val

        def master():
            i = 0
            while 1:
                yield clk.posedge
                msig.next = uprange[i + 1]
                assert msig.val == uprange[i]
                shared.t = now()
                i += 1

        def slave(ssig):
            """ Double-check proper operation """
            i = 0
            while 1:
                yield clk.posedge
                ssig.next = msig.val
                assert ssig.val == uprange[i - 1]
                i += 1

        def deltaSlave(dsig):
            """ Expect delta cycle races """
            i = 0
            while 1:
                yield deltaClk.posedge
                dsig.next = msig.val
                assert now() == shared.t
                assert dsig.val == uprange[i]
                i += 1

        return (slave(ssig[1]), deltaSlave(dsig[1]),
                master(), clkGen(), deltaClkGen(),
                slave(ssig[0]), deltaSlave(dsig[0]))

    def testDeltaCycleRace(self):
        """ Check delta cycle races """
        bench = self.bench()
        Simulation(bench).run(quiet=QUIET)


class DelayLine(TestCase):

    """ Check that delay lines work properly """

    def bench(self):
        uprange = range(500)
        sig_Z = [Signal(uprange[-i]) for i in range(7)]
        clk = Signal(0)

        def clkGen():
            for i in uprange[:-1]:
                yield delay(10)
                clk.next = 1
                yield delay(10)
                clk.next = 0

        def delayElement(n, i):
                sig_Z[n].next = sig_Z[n - 1].val
                assert sig_Z[n].val == uprange[i - n]

        def stage(n):
            i = 0
            while 1:
                yield clk.posedge
                delayElement(n, i)
                i += 1

        def stage012():
            i = 0
            while 1:
                yield clk.posedge
                delayElement(1, i)
                sig_Z[0].next = uprange[i + 1]
                delayElement(2, i)
                i += 1

        return [stage(6), stage(4), clkGen(), stage(3), stage012(), stage(5)]

    def testZeroDelay(self):
        """ Zero delay behavior """
        bench = self.bench()
        Simulation(bench).run(quiet=QUIET)


def initSignal(waveform):
    __, val, sigdelay = waveform[0]
    if sigdelay:
        return Signal(val=val, delay=sigdelay)
    else:
        return Signal(val=val)


def isPosedge(oldval, val):
    return not oldval and val


def isNegedge(oldval, val):
    return oldval and not val


def isEvent(oldval, val):
    return oldval != val


def isEdge(oldval, val):
    return isPosedge(oldval, val) or isNegedge(oldval, val)


def getExpectedTimes(waveform, eventCheck):
    interval, val, sigdelay = waveform[0]
    # print waveform[0]
    expected = []
    time = interval
    oldval = val
    i = 1
    while i < len(waveform):
        interval, val, sigdelay = waveform[i]
        # print waveform[i]
        time += interval
        # check future events within inertial delay interval
        j = i + 1
        inctime = 0
        while j < len(waveform) and inctime + waveform[j][0] < sigdelay:
            inctime += waveform[j][0]
            newval = waveform[j][1]
            newsigdelay = waveform[j][2]
            if newval != val:  # cancel event
                break
            else:  # same vals
                if inctime + newsigdelay < sigdelay:
                    # special case: there is a later event, with same val,
                    # but smaller delay: presumably, this should win,
                    # so cancel the present one
                    break
            j += 1
        else:  # if event was not cancelled by a break
            if eventCheck(oldval, val):
                expected.append(time + sigdelay)
                # print expected[-1]
            oldval = val
        i += 1
    # print expected
    return expected


class Waveform(TestCase):

    """ Test of all sorts of event response in a waveform """

    waveform = []
    duration = 0
    sigdelay = 0
    for i in range(2000):
        interval = randrange(0, 150)
        val = randrange(0, 4)
        waveform.append((interval, val, sigdelay))
        duration = interval + duration

    def stimulus(self):
        for interval, val, sigdelay in self.waveform:
            yield delay(interval)
            self.sig.next = val
            if sigdelay:
                self.sig.delay = sigdelay

    def response(self, clause, expected):
        assert len(expected) > 100  # we should test something
        i = 0
        while 1:
            yield clause
            assert now() == expected[i]
            i += 1

    def setUp(self):
        self.sig = initSignal(self.waveform)

    def runSim(self, sim):
        sim.run(quiet=QUIET)

    def testPosedge(self):
        """ Posedge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isPosedge)
        response = self.response(clause=s.posedge, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testNegedge(self):
        """ Negedge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isNegedge)
        response = self.response(clause=s.negedge, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testEdge(self):
        """ Edge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isEdge)
        response = self.response(clause=(s.negedge, s.posedge),
                                 expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testEvent(self):
        """ Event waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isEvent)
        # print expected
        response = self.response(clause=s, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testRedundantEvents(self):
        """ Redundant event waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isEvent)
        response = self.response(clause=(s,) * 6, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testRedundantEventAndEdges(self):
        """ Redundant edge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isEvent)
        response = self.response(clause=(s, s.negedge, s.posedge),
                                 expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testRedundantPosedges(self):
        """ Redundant posedge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isPosedge)
        response = self.response(clause=(s.posedge,) * 3, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()

    def testRedundantNegedges(self):
        """ Redundant negedge waveform test """
        s = self.sig
        stimulus = self.stimulus()
        expected = getExpectedTimes(self.waveform, isNegedge)
        response = self.response(clause=(s.negedge,) * 9, expected=expected)
        self.runSim(Simulation(stimulus, response))
        assert self.duration <= now()


class WaveformSigDelay(Waveform):

    """ Repeat waveform in tests with a delayed signal """

    waveform = []
    duration = 0
    sigdelay = 0
    for i in range(2000):
        interval = randrange(20, 150)
        val = randrange(0, 4)
        sigdelay = randrange(1, 20)
        waveform.append((interval, val, sigdelay))
        duration += interval


class WaveformInertialDelay(Waveform):

    """ Repeat waveform tests to check inertial delay """

    waveform = []
    duration = 0
    sigdelay = 0
    for i in range(2000):
        interval = randrange(3, 10)
        val = randrange(0, 3)
        sigdelay = randrange(1, 5)
        waveform.append((interval, val, sigdelay))
        duration += interval


class WaveformInertialDelayStress(Waveform):

    """ Repeat waveform tests to stress inertial delay """

    waveform = []
    duration = 0
    sigdelay = 0
    for i in range(2000):
        interval = randrange(1, 3)
        val = randrange(0, 3)
        sigdelay = randrange(1, 3)
        waveform.append((interval, val, sigdelay))
        duration += interval


class SimulationRunMethod(Waveform):

    """ Basic test of run method of Simulation object """

    def runSim(self, sim):
        duration = randrange(1, 300)
        while sim.run(duration, quiet=QUIET):
            duration = randrange(1, 300)


class TimeZeroEvents(TestCase):

    """ Check events at time 0 """

    def bench(self, sig, nextval, clause, timeout=1):

        def stimulus():
            sig.next = nextval
            yield delay(10)

        def response():
            yield clause, delay(timeout)
            assert now() == 0
            assert sig.val == nextval

        return [stimulus(), response()]

    def testEvent(self):
        """ Event at time 0 """
        s = Signal(0)
        testBench = self.bench(sig=s, nextval=1, clause=s)
        Simulation(testBench).run(quiet=QUIET)

    def testPosedge(self):
        """ Posedge at time 0 """
        s = Signal(0)
        testBench = self.bench(sig=s, nextval=1, clause=s.posedge)
        Simulation(testBench).run(quiet=QUIET)

    def testNegedge(self):
        """ Negedge at time 0 """
        s = Signal(1)
        testBench = self.bench(sig=s, nextval=0, clause=s.negedge)
        Simulation(testBench).run(quiet=QUIET)



import os
path = os.path

import random
random.seed(2)

from myhdl import (block, Signal, intbv, delay, always, instance)

from myhdl.conversion import verify

ACTIVE_LOW, INACTIVE_HIGH = bool(0), bool(1)


@block
def incRef(count, enable, clock, reset, n):
    """ Incrementer with enable.
    
    count -- output
    enable -- control input, increment when 1
    clock -- clock input
    reset -- asynchronous reset input
    n -- counter max value
    """

    @instance
    def comb():
        while 1:
            yield clock.posedge, reset.negedge
            if reset == ACTIVE_LOW:
                count.next = 0
            else:
                if enable:
                    count.next = (count + 1) % n

    return comb


@block
def inc(count, enable, clock, reset, n):

    """ Incrementer with enable.
    
    count -- output
    enable -- control input, increment when 1
    clock -- clock input
    reset -- asynchronous reset input
    n -- counter max value
    
    """

    @always(clock.posedge, reset.negedge)
    def incProcess():
        if reset == 0:
            count.next = 0
        else:
            if enable:
                count.next = (count + 1) % n

    return incProcess


@block
def inc2(count, enable, clock, reset, n):

    @always(clock.posedge, reset.negedge)
    def incProcess():
        if reset == ACTIVE_LOW:
            count.next = 0
        else:
            if enable:
                if count == n - 1:
                    count.next = 0
                else:
                    count.next = count + 1

    return incProcess


@block
def incFunc(count, enable, clock, reset, n):

    def incFuncFunc(cnt, enable):
        count_next = intbv(0, min=0, max=n)
        count_next[:] = cnt
        if enable:
            count_next[:] = cnt + 1
        return count_next

    @always(clock.posedge, reset.negedge)
    def incFuncGen():
        if reset == ACTIVE_LOW:
            count.next = 0
        else:
            count.next = incFuncFunc(count, enable)

    return incFuncGen


@block
def incTask(count, enable, clock, reset, n):

    def incTaskFunc(cnt, enable, n):
        if enable:
            cnt[:] = (cnt + 1) % n

    @instance
    def incTaskGen():
        cnt = intbv(0)[8:]
        while 1:
            yield clock.posedge, reset.negedge
            if reset == ACTIVE_LOW:
                cnt[:] = 0
                count.next = 0
            else:
                # print count
                incTaskFunc(cnt, enable, n)
                count.next = cnt

    return incTaskGen


@block
def incTaskFreeVar(count, enable, clock, reset, n):

    def incTaskFunc():
        if enable:
            count.next = (count + 1) % n

    @always(clock.posedge, reset.negedge)
    def incTaskGen():
        if reset == ACTIVE_LOW:
            count.next = 0
        else:
            # print count
            incTaskFunc()

    return incTaskGen


@block
def IncBench(inc):

    NR_CYCLES = 201

    m = 8
    n = 2 ** m

    count = Signal(intbv(0)[m:])
    enable = Signal(bool(0))
    clock, reset = [Signal(bool(0)) for __ in range(2)]

    inc_inst = inc(count, enable, clock, reset, n=n)

    @instance
    def clockgen():
        clock.next = 1
        for dummy in range(NR_CYCLES):
            yield delay(10)
            clock.next = not clock

    @instance
    def monitor():
        reset.next = 0
        enable.next = 1
        yield clock.negedge
        reset.next = 1
        yield clock.negedge
        while True:
            yield clock.negedge
            print(count)

    return inc_inst, clockgen, monitor


def test_incReg():
    assert verify(IncBench(incRef)) == 0


def test_inc():
    assert verify(IncBench(inc)) == 0


def test_inc2():
    assert verify(IncBench(inc2)) == 0


def testIncTask():
    assert verify(IncBench(incTask)) == 0


def testIncFunc():
    assert verify(IncBench(incFunc)) == 0




import os

from myhdl import Cosimulation

cmd = 'vsim -c -quiet -pli ../myhdl_vpi.dll -do cosim.do dut_bin2gray'

def bin2gray(B, G, width):
    os.system('vlog -quiet +define+width=%s ../../test/verilog/bin2gray.v' % (width))
    os.system('vlog -quiet +define+width=%s ../../test/verilog/dut_bin2gray.v' % (width))

    return Cosimulation(cmd, B=B, G=G)



from myhdl import (block, Signal, ResetSignal, intbv, always_seq,
                   instance, delay, StopSimulation,)


class MyIntf(object):

    def __init__(self):
        self.x = Signal(intbv(2, min=0, max=16))
        self.y = Signal(intbv(3, min=0, max=18))


@block
def one_level(clock, reset, ia, ib):

    @always_seq(clock.posedge, reset=reset)
    def rtl():
        ia.x.next = ib.x + 1
        ia.y.next = ib.y + 1

    return rtl


@block
def two_level(clock, reset, ia, ib):

    ic, ie = MyIntf(), MyIntf()
    one_inst = one_level(clock, reset, ic, ie)

    @always_seq(clock.posedge, reset=reset)
    def rtl():
        ia.x.next = ib.x + ic.x
        ia.y.next = ib.y + ic.y

    return one_inst, rtl


@block
def c_testbench_one():
    clock = Signal(bool(0))
    reset = ResetSignal(0, active=0, isasync=True)
    ia, ib = MyIntf(), MyIntf()

    tb_dut = one_level(clock, reset, ia, ib)

    @instance
    def tb_clk():
        clock.next = False
        yield delay(10)
        while True:
            clock.next = not clock
            yield delay(10)

    @instance
    def tb_stim():
        reset.next = False
        yield delay(17)
        reset.next = True
        yield delay(17)
        for ii in range(7):
            yield clock.posedge
        assert ia.x == 3
        assert ia.y == 4
        print("%d %d %d %d" % (ia.x, ia.y, ib.x, ib.y))
        raise StopSimulation

    return tb_dut, tb_clk, tb_stim


@block
def c_testbench_two():
    clock = Signal(bool(0))
    reset = ResetSignal(0, active=0, isasync=True)
    ia, ib = MyIntf(), MyIntf()

    tb_dut = two_level(clock, reset, ia, ib)

    @instance
    def tb_clk():
        clock.next = False
        yield delay(10)
        while True:
            clock.next = not clock
            yield delay(10)

    @instance
    def tb_stim():
        reset.next = False
        yield delay(17)
        reset.next = True
        yield delay(17)
        for ii in range(7):
            yield clock.posedge
        assert ia.x == 5
        assert ia.y == 7
        print("%d %d %d %d" % (ia.x, ia.y, ib.x, ib.y))
        raise StopSimulation

    return tb_dut, tb_clk, tb_stim


def test_one_level_analyze():
    clock = Signal(bool(0))
    reset = ResetSignal(0, active=0, isasync=True)
    ia, ib = MyIntf(), MyIntf()
    inst = one_level(clock, reset, ia, ib)
    assert inst.analyze_convert() == 0


def test_one_level_verify():
    inst = c_testbench_one()
    assert inst.verify_convert() == 0


def test_two_level_analyze():
    clock = Signal(bool(0))
    reset = ResetSignal(0, active=0, isasync=True)
    ia, ib = MyIntf(), MyIntf()
    inst = two_level(clock, reset, ia, ib)
    assert inst.analyze_convert() == 0


def test_two_level_verify():
    inst = c_testbench_two()
    assert inst.verify_convert() == 0



from myhdl import (block, Signal, intbv, delay, always_comb, instance,
                   TristateSignal, ConcatSignal, conversion, StopSimulation,
                   toVerilog, toVHDL)


@block
def bench_SliceSignal():

    s = Signal(intbv(0)[8:])
    a, b, c = s(7), s(5), s(0)
    d, e, f, g = s(8, 5), s(6, 3), s(8, 0), s(4, 3)
    N = len(s)

    @instance
    def check():
        ''' Running bench_SliceSignal '''
        for i in range(N):
            s.next = i
            yield delay(10)
            print(int(a))
            print(int(b))
            print(int(c))
            print(d)
            print(e)
            print(f)
            print(g)

    return check


def test_SliceSignal():
    assert conversion.verify(bench_SliceSignal()) == 0


@block
def bench_RecurseShadowSignal():
    s = Signal(intbv(0)[16:])

    # take some recursive ShadowSignals
    su = s(16, 8)
    sl = s(8, 0)
    suu = su(8, 4)
    sul = su(4, 0)
    slu = sl(8, 4)
    sll = sl(4, 0)

    @instance
    def check():
        for i in range(25):
            s.next = 0x4567 + i * 39
            yield delay(10)
            print(s)
            print(int(su))
            print(int(sl))
            print(suu)
            print(sul)
            print(slu)
            print(sll)

    return check


def test_RecurseShadowSignal():
    assert conversion.verify(bench_RecurseShadowSignal()) == 0


@block
def bench_ConcatSignal():

    a = Signal(intbv(0)[5:])
    b = Signal(bool(0))
    c = Signal(intbv(0)[3:])
    d = Signal(intbv(0)[4:])

    s = ConcatSignal(a, b, c, d)

    I_max = 2 ** len(a)
    J_max = 2 ** len(b)
    K_max = 2 ** len(c)
    M_max = 2 ** len(d)

    @instance
    def check():
        for i in range(I_max):
            for j in range(J_max):
                for k in range(K_max):
                    for m in range(M_max):
                        a.next = i
                        b.next = j
                        c.next = k
                        d.next = m
                        yield delay(10)
                        print(s)

    return check


def test_ConcatSignal():
    assert conversion.verify(bench_ConcatSignal()) == 0


@block
def bench_ConcatSignalWithConsts():

    a = Signal(intbv(0)[5:])
    b = Signal(bool(0))
    c = Signal(intbv(0)[3:])
    d = Signal(intbv(0)[4:])
    e = Signal(intbv(0)[1:])

    c1 = "10"
    c2 = intbv(3)[3:]
    c3 = '0'
    c4 = bool(1)
    c5 = intbv(42)[8:]  # with leading zeroes

    s = ConcatSignal(c1, a, c2, b, c3, c, c4, d, c5, e)

    I_max = 2 ** len(a)
    J_max = 2 ** len(b)
    K_max = 2 ** len(c)
    M_max = 2 ** len(d)

    @instance
    def check():
        for i in range(I_max):
            for j in range(J_max):
                for k in range(K_max):
                    for m in range(M_max):
                        for n in range(2 ** len(e)):
                            a.next = i
                            b.next = j
                            c.next = k
                            d.next = m
                            e.next = n
                            yield delay(10)
                            print(s)

    return check


def test_ConcatSignalWithConsts():
    assert conversion.verify(bench_ConcatSignalWithConsts()) == 0


@block
def bench_ConcatSignalUndriven():

    n_sigs = 32

    sig_list = [Signal(False) for __ in range(n_sigs)]
    concat_sig = ConcatSignal(*reversed(sig_list))

    @instance
    def check():

        yield delay(10)
        print(concat_sig)

    return check


def test_ConcatSignalUndriven():
    assert conversion.verify(bench_ConcatSignalUndriven()) == 0


@block
def bench_TristateSignal():
    s = TristateSignal(intbv(0)[8:])
    a = s.driver()
    b = s.driver()
    c = s.driver()

    @instance
    def check():
        a.next = None
        b.next = None
        c.next = None
        yield delay(10)
        # print s
        a.next = 1
        yield delay(10)
        print(s)
        a.next = None
        b.next = 122
        yield delay(10)
        print(s)
        b.next = None
        c.next = 233
        yield delay(10)
        print(s)
        c.next = None
        yield delay(10)
        # print s

    return check


def test_TristateSignal():
    assert conversion.verify(bench_TristateSignal()) == 0


@block
def permute(x, a, mapping):

    p = [a(m) for m in mapping]

    q = ConcatSignal(*p)

    @always_comb
    def assign():
        x.next = q

    return assign


@block
def bench_permute(conv=False):

    x = Signal(intbv(0)[3:])
    a = Signal(intbv(0)[3:])
    mapping = (0, 2, 1)

    if conv:
        dut = conv(permute(x, a, mapping))
    else:
        dut = permute(x, a, mapping)

    @instance
    def stimulus():
        for i in range(2 ** len(a)):
            a.next = i
            yield delay(10)
            print("%d %d" % (x, a))
            assert x[2] == a[0]
            assert x[1] == a[2]
            assert x[0] == a[1]
        raise StopSimulation()

    return dut, stimulus


def test_permute():
    assert conversion.verify(bench_permute()) == 0


bench_permute(toVHDL)
bench_permute(toVerilog)



from myhdl import block, Signal

from ClkDriver import ClkDriver
from Hello import Hello


@block
def Greetings():

    clk1 = Signal(0)
    clk2 = Signal(0)

    clkdriver_1 = ClkDriver(clk1)  # positional and default association
    clkdriver_2 = ClkDriver(clk=clk2, period=19)  # named association
    hello_1 = Hello(clk=clk1)  # named and default association
    hello_2 = Hello(to="MyHDL", clk=clk2)  # named association

    return clkdriver_1, clkdriver_2, hello_1, hello_2


inst = Greetings()
inst.run_sim(50)



""" myhdl's distribution and installation script. """

import ast
import fnmatch
import re
import os
import sys

from collections import defaultdict

if sys.version_info < (3, 7):
    raise RuntimeError("Python version 3.7+ required.")


# Prefer setuptools over distutils
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('myhdl/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

data_root = 'share/myhdl'
cosim_data = defaultdict(list)
for base, dir, files in os.walk('cosimulation'):
    for pat in ('*.c', 'Makefile*', '*.py', '*.v', '*.txt'):
        good = fnmatch.filter(files, pat)
        if good:
            cosim_data[base].extend(os.path.join(base, f) for f in good)

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="myhdl",
    version=version,
    description="Python as a Hardware Description Language",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    author="Jan Decaluwe",
    author_email="jan@jandecaluwe.com",
    url="http://www.myhdl.org",
    packages=['myhdl', 'myhdl.conversion'],
    data_files=[(os.path.join(data_root, k), v) for k, v in cosim_data.items()],
    license="LGPL",
    platforms='any',
    keywords="HDL ASIC FPGA hardware design",
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3 :: Only',
    ]
)



import myhdl
from myhdl import *
from myhdl.conversion import analyze


def dff(q, d, clk):

    @always(clk.posedge)
    def comb():
        q.next = d

    return comb


from random import randrange


def test_dff():

    q, d, clk = [Signal(bool(0)) for i in range(3)]

    dff_inst = dff(q, d, clk)

    @always(delay(10))
    def clkgen():
        clk.next = not clk

    @always(clk.negedge)
    def stimulus():
        d.next = randrange(2)

    return dff_inst, clkgen, stimulus


def simulate(timesteps):
    traceSignals.timescale = "1ps"
    tb = traceSignals(test_dff)
    sim = Simulation(tb)
    sim.run(timesteps)
    sim.quit()


simulate(2000)


def convert():
    q, d, clk = [Signal(bool(0)) for i in range(3)]
    toVerilog(dff, q, d, clk)
    analyze(dff, q, d, clk)


convert()




#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2011 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.

#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" Module with the modbv class """
from ._intbv import intbv


class modbv(intbv):
    __slots__ = []

    def _handleBounds(self):
        lo, hi, val = self._min, self._max, self._val
        if lo is not None:
            if val < lo or val >= hi:
                self._val = (val - lo) % (hi - lo) + lo

    def __repr__(self):
        return "modbv(" + repr(self._val) + ")"

    # indexing and slicing methods
    # dedicated for modbv to support "declaration by slicing"

    def __getitem__(self, key):
        if isinstance(key, slice):
            i, j = key.start, key.stop
            if j is None:  # default
                j = 0
            j = int(j)
            if j < 0:
                raise ValueError("modbv[i:j] requires j >= 0\n"
                                 "            j == %s" % j)
            if i is None:  # default
                return modbv(self._val >> j)
            i = int(i)
            if i <= j:
                raise ValueError("modbv[i:j] requires i > j\n"
                                 "            i, j == %s, %s" % (i, j))
            res = modbv((self._val & (1 << i) - 1) >> j, _nrbits=i - j)
            return res
        else:
            i = int(key)
            res = bool((self._val >> i) & 0x1)
            return res



from myhdl import (block, Signal, intbv, always, concat, toVerilog, toVHDL)
from myhdl.conversion import analyze


@block
def issue_117(clk, sdi, pdo, sel, const=False):
    assert isinstance(const, (bool, intbv))
    delay_reg = Signal(intbv(0)[8:])
    rlen = len(pdo)
    plen = 1 if isinstance(const, bool) else len(const)

    @always(clk.posedge)
    def rtl():
        if sel == 0:
            delay_reg.next = concat(const, delay_reg[rlen - plen - 1:1], sdi)
        elif sel == 1:
            delay_reg.next = concat(delay_reg[rlen - 1:plen + 1], const, sdi)
        elif sel == 2:
            delay_reg.next = concat(delay_reg[rlen - 1:plen + 1], sdi, const)
            pdo.next = delay_reg

    return rtl


def test_issue_117_1():
    clk, sdi = [Signal(bool(0)) for _ in range(2)]
    pdo = Signal(intbv(0)[8:])
    sel = Signal(intbv(0, min=0, max=3))
    toVHDL.name = toVerilog.name = 'issue_117_1'
    assert issue_117(clk, sdi, pdo, sel, const=bool(0)).analyze_convert() == 0


def test_issue_117_2():
    clk, sdi = [Signal(bool(0)) for _ in range(2)]
    pdo = Signal(intbv(0)[8:])
    sel = Signal(intbv(0, min=0, max=3))
    toVHDL.name = toVerilog.name = 'issue_117_2'
    assert issue_117(clk, sdi, pdo, sel, const=False).analyze_convert() == 0


def test_issue_117_3():
    clk, sdi = [Signal(bool(0)) for _ in range(2)]
    pdo = Signal(intbv(0)[8:])
    sel = Signal(intbv(0, min=0, max=3))
    toVHDL.name = toVerilog.name = 'issue_117_3'
    assert issue_117(clk, sdi, pdo, sel, const=intbv(0)[1:]).analyze_convert() == 0


if __name__ == '__main__':
    analyze.simulator = 'vlog'
    test_issue_117_1()




from myhdl import (block, Signal, enum, intbv, delay, instance,
                   instances, always_comb, StopSimulation)

bitwise_op = enum('BW_AND', 'BW_ANDN', 'BW_OR', 'BW_XOR')


def bitwise(a, b, op):
    r = intbv(0)[8:]
    if op == bitwise_op.BW_AND:
        r[:] = a & b
    elif op == bitwise_op.BW_ANDN:
        r[:] = (~a) & b
    elif op == bitwise_op.BW_OR:
        r[:] = a | b
    elif op == bitwise_op.BW_XOR:
        r[:] = a ^ b
    return r


@block
def LogicUnit(a, b, c, op):

    @always_comb
    def operate():
        c.next = bitwise(a, b, op)

    return operate


@block
def bench_enum():
    clock = Signal(False)
    a, b, c = [Signal(intbv(0)[8:]) for _ in range(3)]
    op = Signal(bitwise_op.BW_AND)

    logic_unit = LogicUnit(a=a, b=b, c=c, op=op)

    @instance
    def clockgen():
        clock.next = 1
        while 1:
            yield delay(10)
            clock.next = not clock

    @instance
    def stimulus():
        a.next = 0xaa
        b.next = 0x55
        yield clock.posedge
        print('a=%s b=%s' % (a, b))

        op.next = bitwise_op.BW_AND
        yield clock.posedge
        print(c)

        op.next = bitwise_op.BW_ANDN
        yield clock.posedge
        print(c)

        op.next = bitwise_op.BW_OR
        yield clock.posedge
        print(c)

        op.next = bitwise_op.BW_XOR
        yield clock.posedge
        print(c)

        raise StopSimulation

    return instances()


def test_enum():
    assert bench_enum().verify_convert() == 0




from random import randrange

import myhdl
from myhdl import *

from TimeCount import TimeCount

LOW, HIGH = bool(0), bool(1)

MAX_COUNT = 6 * 10 * 10
PERIOD = 10

def bench():

    """ Unit test for time counter. """

    tens, ones, tenths = [Signal(intbv(0)[4:]) for i in range(3)]
    startstop, reset, clock = [Signal(LOW) for i in range(3)]

    dut = TimeCount(tens, ones, tenths, startstop, reset, clock)
    
    count = Signal(0)
    counting = Signal(False)

    @always(delay(PERIOD//2))
    def clkgen():
        clock.next = not clock

##     @always(reset.posedge)
##     def clear():
##         counting.next = False
##         count.next = 0
            
##     @always(startstop.posedge)
##     def go():
##         counting.next = not counting

    @always(startstop.posedge, reset.posedge)
    def action():
        if reset:
            counting.next = False
            count.next = 0
        else:
            counting.next = not counting      

    @always(clock.posedge)
    def counter():
        if counting:
            count.next = (count + 1) % MAX_COUNT
            
    @always(clock.negedge)
    def monitor():
        assert ((tens*100) + (ones*10) + tenths) == count

    @instance
    def stimulus():
        for maxInterval in (100*PERIOD, 2*MAX_COUNT*PERIOD):
            for sig in (reset, startstop,
                        reset, startstop, startstop,
                        reset, startstop, startstop, startstop,
                        reset, startstop, reset, startstop, startstop, startstop):
               yield delay(randrange(10*PERIOD, maxInterval))
               yield clock.negedge # sync to avoid race condition
               sig.next = HIGH
               yield delay(100)
               sig.next = LOW
        raise StopSimulation
  
    return dut, clkgen, action, counter, monitor, stimulus


def test_bench():
    sim = Simulation(bench())
    sim.run()

def convert():
    tens, ones, tenths = [Signal(intbv(0)[4:]) for i in range(3)]
    startstop, reset, clock = [Signal(LOW) for i in range(3)]

    toVerilog(TimeCount, tens, ones, tenths, startstop, reset, clock)
    toVHDL(TimeCount, tens, ones, tenths, startstop, reset, clock)

convert()



import os

from myhdl import Cosimulation

cmd = "cver -q +loadvpi=../myhdl_vpi:vpi_compat_bootstrap +define+width=%s " + \
      "../../test/verilog/bin2gray.v " + \
      "../../test/verilog/dut_bin2gray.v "
       
def bin2gray(B, G, width):
    return Cosimulation(cmd % width, B=B, G=G)
               




import os
path = os.path
import unittest
from unittest import TestCase
import random
from random import randrange
random.seed(2)

from myhdl import (block, Signal, intbv, delay, instance, StopSimulation,)
from myhdl._Simulation import Simulation

from .util import setupCosimulation

ACTIVE_LOW, INACTIVE_HIGH = 0, 1


@block
def behRef(count, enable, clock, reset, n):

    @instance
    def comb():
        while 1:
            if reset == ACTIVE_LOW:
                yield reset.posedge
            for i in range(20):
                yield clock.posedge
                if enable:
                    count.next = i
            j = 1
            while j < 25:
                if enable:
                    yield clock.posedge
                yield clock.posedge
                count.next = 2 * j
                j += 1

    return comb


objfile = "beh_inst.o"
analyze_cmd = "iverilog -o %s beh_inst.v tb_beh_inst.v" % objfile
simulate_cmd = "vvp -m ../../../cosimulation/icarus/myhdl.vpi %s" % objfile


def beh_v(name, count, enable, clock, reset):
    return setupCosimulation(**locals())


class TestBeh(TestCase):

    def clockGen(self, clock):
        while 1:
            yield delay(10)
            clock.next = not clock

    def stimulus(self, enable, clock, reset):
        reset.next = INACTIVE_HIGH
        yield clock.negedge
        reset.next = ACTIVE_LOW
        yield clock.negedge
        reset.next = INACTIVE_HIGH
        for dummy in range(1000):
            enable.next = 1
            yield clock.negedge
        for dummy in range(1000):
            enable.next = min(1, randrange(5))
            yield clock.negedge
        raise StopSimulation

    def check(self, count, count_v, enable, clock, reset, n):
        yield reset.posedge
        self.assertEqual(count, count_v)
        while 1:
            yield clock.posedge
            yield delay(1)
            # print "%d count %s count_v %s" % (now(), count, count_v)
            self.assertEqual(count, count_v)

    def bench(self, beh):

        m = 8
        n = 2 ** m

        count = Signal(intbv(0)[m:])
        count_v = Signal(intbv(0)[m:])
        enable = Signal(bool(0))
        clock, reset = [Signal(bool()) for __ in range(2)]

        beh_inst = beh(count, enable, clock, reset, n=n).convert(hdl='Verilog')
        # beh_inst = beh(count, enable, clock, reset, n=n)
        beh_inst_v = beh_v(beh.__name__, count_v, enable, clock, reset)
        clk_1 = self.clockGen(clock)
        st_1 = self.stimulus(enable, clock, reset)
        ch_1 = self.check(count, count_v, enable, clock, reset, n=n)

        sim = Simulation(beh_inst, beh_inst_v, clk_1, st_1, ch_1)
        # sim = Simulation(beh_inst,  clk_1, st_1, ch_1)
        return sim

    def testBehRef(self):
        sim = self.bench(behRef)
        sim.run(quiet=1)


if __name__ == '__main__':
    unittest.main()




import os
path = os.path
import unittest
from unittest import TestCase

from myhdl import (block, Signal, intbv, delay, always, always_comb,
                   instance, StopSimulation)
from myhdl._Simulation import Simulation

from .util import setupCosimulation


@block
def ram(dout, din, addr, we, clk, depth=128):
    """ Simple ram model """

    @instance
    def comb():
        mem = [intbv(0)[8:] for dummy in range(depth)]
        # ad = 1
        while 1:
            yield clk.posedge
            if we:
                mem[int(addr)][:] = din
            dout.next = mem[int(addr)]

    return comb


@block
def ram_clocked(dout, din, addr, we, clk, depth=128):
    """ Ram model """

    mem = [Signal(intbv(0)[8:]) for __ in range(depth)]

    @instance
    def access():
        while 1:
            yield clk.posedge
            if we:
                mem[int(addr)].next = din
            dout.next = mem[int(addr)]

    return access


@block
def ram_deco1(dout, din, addr, we, clk, depth=128):
    """  Ram model """

    mem = [Signal(intbv(0)[8:]) for __ in range(depth)]

    @instance
    def write():
        while 1:
            yield clk.posedge
            if we:
                mem[int(addr)].next = din

    @always_comb
    def read():
        dout.next = mem[int(addr)]

    return write, read


@block
def ram_deco2(dout, din, addr, we, clk, depth=128):
    """  Ram model """

    mem = [Signal(intbv(0)[8:]) for __ in range(depth)]

    @always(clk.posedge)
    def write():
        if we:
            mem[int(addr)].next = din

    @always_comb
    def read():
        dout.next = mem[int(addr)]

    return write, read


@block
def ram2(dout, din, addr, we, clk, depth=128):

    # memL = [intbv(0,min=dout._min,max=dout._max) for i in range(depth)]
    memL = [Signal(intbv()[len(dout):]) for __ in range(depth)]

    @instance
    def wrLogic():
        while 1:
            yield clk.posedge
            if we:
                memL[int(addr)].next = din

    @instance
    def rdLogic():
        while 1:
            yield clk.posedge
            dout.next = memL[int(addr)]

    return wrLogic, rdLogic


@block
def ram_v(name, dout, din, addr, we, clk, depth=4):
    return setupCosimulation(**locals())


class TestMemory(TestCase):

    def bench(self, ram, depth=128):

        dout = Signal(intbv(0)[8:])
        dout_v = Signal(intbv(0)[8:])
        din = Signal(intbv(0)[8:])
        addr = Signal(intbv(0)[7:])
        we = Signal(bool(0))
        clk = Signal(bool(0))

        mem_inst = ram(dout, din, addr, we, clk, depth).convert(hdl='Verilog')
        mem_v_inst = ram_v(ram.__name__, dout_v, din, addr, we, clk, depth)

        def stimulus():
            for i in range(depth):
                din.next = i
                addr.next = i
                we.next = True
                yield clk.negedge
            we.next = False
            for i in range(depth):
                addr.next = i
                yield clk.negedge
                yield clk.posedge
                yield delay(1)
                # print dout
                # print dout_v
                self.assertEqual(dout, i)
                # self.assertEqual(dout, dout_v)
            raise StopSimulation()

        def clkgen():
            while 1:
                yield delay(10)
                clk.next = not clk

        return clkgen(), stimulus(), mem_inst, mem_v_inst

    def test1(self):
        sim = self.bench(ram)
        Simulation(sim).run()

    def test2(self):
        sim = self.bench(ram2)
        Simulation(sim).run()

    def testram_clocked(self):
        sim = self.bench(ram_clocked)
        Simulation(sim).run()

    def testram_deco1(self):
        sim = self.bench(ram_deco1)
        Simulation(sim).run()

    def testram_deco2(self):
        sim = self.bench(ram_deco2)
        Simulation(sim).run()


if __name__ == '__main__':
    unittest.main()




#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.

#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" Run the unit tests for enum """
import copy
import random

import pytest

from myhdl import enum

random.seed(1)  # random, but deterministic

t_State = enum("SEARCH", "CONFIRM", "SYNC")
t_Homograph = enum("SEARCH", "CONFIRM", "SYNC")
t_incomplete = enum("SEARCH", "CONFIRM")


class TestEnum:

    def testUniqueLiterals(self):
        with pytest.raises(ValueError):
            t_State = enum("SEARCH", "CONFIRM", "SEARCH")

    def testWrongAttr(self):
        with pytest.raises(AttributeError):
            t_State.TYPO

    def testAttrAssign(self):
        with pytest.raises(AttributeError):
            t_State.SEARCH = 4

    def testWrongAttrAssign(self):
        with pytest.raises(AttributeError):
            t_State.TYPO = 4

    def testHomograph(self):
        assert t_State is not t_Homograph

    def testHomographLiteral(self):
        assert t_State.SEARCH is not t_Homograph.SEARCH

    def testItemCopy(self):
        e = copy.deepcopy(t_State.SEARCH)
        assert e == t_State.SEARCH
        assert e != t_State.CONFIRM

# # Adding test coverage for encoding in enum

    def testItemNotDeepCopy(self):
        e = copy.copy(t_State.SEARCH)
        assert e == t_State.SEARCH
        assert e != t_State.CONFIRM

    def testWrongEncoding(self):

        def logic1(encoding):
            t_State = enum("SEARCH", "CONFIRM", "SYNC", encoding=encoding)
            with pytest.raises(ValueError):
                logic1(encoding)

    def testNotStringtype(self):
        with pytest.raises(TypeError):
            t_State = enum("SEARCH", 1, "SYNC")

    def testEnumLength(self):
        l = len(t_State)
        assert l == len(t_State)




#!/usr/bin/env python
#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.

#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" Run the intbv.signed() unit tests. """
from random import randrange

from myhdl import Signal, intbv, concat


class TestIntbvSigned:
    '''Test cases to verify the intbv.signed() member function'''

    def testPlainIntbvInstance(self):
        '''Test a plain intbv instance with .signed() 

        ----+----+----+----+----+----+----+----
           -3   -2   -1    0    1    2    3

                          min  max
                               min  max
                     min       max
                     min            max
                min            max
                min       max
                min  max
              neither min nor max is set
              only max is set
              only min is set

        '''
        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # in the following cases the .signed() function should classify the
        # value of the intbv instance as unsigned and return the 2's
        # complement value of the bits as specified by _nrbits.
        #

        # intbv with positive range, pos number, and msb not set, return signed()
        # Expect the number to be returned
        a = intbv(0x3b, min=0, max=0x7c)
        b = a.signed()
        assert b == 0x3b

        # intbv with positive range, pos number, and msb set, return signed()
        # test various bit patterns to see that the 2's complement
        # conversion works correct
        # Expect the number to be converted to a negative number
        a = intbv(7, min=0, max=8)
        b = a.signed()
        assert b == -1

        a = intbv(6, min=0, max=8)
        b = a.signed()
        assert b == -2

        a = intbv(5, min=0, max=8)
        b = a.signed()
        assert b == -3

        # set bit #3 and increase the range so that the set bit is considered
        # the sign bit. Here min = 0
        # Expect to return -4
        a = intbv(4, min=0, max=5)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=0, max=6)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=0, max=7)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=0, max=8)
        b = a.signed()
        assert b == -4

        # here it is not the sign bit anymore
        # Expect the value to be 4
        a = intbv(4, min=0, max=9)
        b = a.signed()
        assert b == 4

        # set bit #3 and increase the range so that the set bit is considered
        # the sign bit. Here min > 0
        # Expect to return -4
        a = intbv(4, min=1, max=5)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=2, max=6)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=3, max=7)
        b = a.signed()
        assert b == -4

        a = intbv(4, min=4, max=8)
        b = a.signed()
        assert b == -4

        # again with min > 0, here it is not the sign bit anymore
        # Expect the value to be 4
        a = intbv(4, min=2, max=9)
        b = a.signed()
        assert b == 4

        # intbv with positive range, value = 0, return signed()
        # Expect the number to be returned
        a = intbv(0, min=0, max=0x8)
        b = a.signed()
        assert b == 0

        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # in these cases the .signed() function should classify the
        # value of the intbv instance as signed and return the value as is
        #

        # intbv without range, pos number, return signed()
        # Expect value to be returned as is
        a = intbv(8)
        b = a.signed()
        assert b == 8

        # intbv without range, neg number, return signed()
        # Expect value to be returned as is
        a = intbv(-8)
        b = a.signed()
        assert b == -8

        # set bit #3 and increase the range that the set bit is actually the
        # msb, but due to the negative range not considered signed
        # Expect to return 4
        a = intbv(4, min=-1, max=5)
        b = a.signed()
        assert b == 4

        a = intbv(4, min=-1, max=6)
        b = a.signed()
        assert b == 4

        a = intbv(4, min=-1, max=7)
        b = a.signed()
        assert b == 4

        a = intbv(4, min=-1, max=8)
        b = a.signed()
        assert b == 4

        # intbv with negative range, pos number, and msb set, return signed()
        # Expect the number to returned as is
        a = intbv(7, min=-1, max=8)
        b = a.signed()
        assert b == 7

        a = intbv(6, min=-1, max=8)
        b = a.signed()
        assert b == 6

        a = intbv(5, min=-1, max=8)
        b = a.signed()
        assert b == 5

        # intbv with symmetric (min = -max) range, pos value, msb set
        # return signed()
        # Expect value returned as is
        a = intbv(4, min=-8, max=8)
        b = a.signed()
        assert b == 4

        # intbv with symmetric (min = -max) range, neg value,
        # return signed()
        # Expect value returned as is
        a = intbv(-4, min=-8, max=8)
        b = a.signed()
        assert b == -4

        # intbv with symmetric (min=-max) range, value = 0,
        # return signed()
        # Expect value returned as is
        a = intbv(0, min=-8, max=8)
        b = a.signed()
        assert b == 0

    def testSlicedSigned(self):
        '''Test a slice with .signed()

        This test can actually be simplified, as a slice will always have
        min=0 and max > min, which will result in an intbv instance that
        will be considered unsigned by the intbv.signed() function.
        '''
        a = intbv(4, min=-8, max=8)
        b = a[4:]
        assert b == 4
        b = a[4:].signed()
        assert b == 4  # msb is not set with a 4 bit slice

        b = a[3:]
        assert b == 4
        b = a[3:].signed()
        assert b == -4  # msb is set with 3 bits sliced

    def testSignedConcat(self):
        '''Test the .signed() function with the concatenate function'''

        # concat 3 bits
        # Expect the signed function to return a negative value
        a = concat(True, True, True).signed()
        assert a == -1

        # concate a 3 bit intbv with msb set and two bits
        # Expect a negative number
        b = concat(intbv(5, min=0, max=8), True, True).signed()
        assert b == -9

    def checkInvariants(self, a):
        """Check invariants of signed operator."""
        W = len(a)
        b = intbv(a.signed())
        if W > 0:
            assert a[W:] == b[W:]
            assert b[:W] == -a[W - 1]
        else:
            assert a == b

    def testRandom(self):
        NRTESTS = 1000
        for L in (10, 1000, 2 ** 32, 2 ** 68):
            for __ in range(NRTESTS):
                lo = randrange(-L, L)
                hi = randrange(lo + 1, 2 * L)
                v = randrange(lo, hi)
                a = intbv(v, min=lo, max=hi)
                self.checkInvariants(a)
                self.checkInvariants(Signal(a))



import os
path = os.path
import random
from random import randrange
random.seed(2)

from myhdl import (block, Signal, intbv, delay, instance, StopSimulation)

NRTESTS = 10


@block
def binaryOps(
              Bitand,
              Bitor,
              Bitxor,
              FloorDiv,
              LeftShift,
              Modulo,
              Mul,
              Pow,
              RightShift,
              Sub,
              Sum,
              EQ,
              NE,
              LT,
              GT,
              LE,
              GE,
              Booland,
              Boolor,
              left, right):

    @instance
    def comb():
        while 1:
            yield left, right
            Bitand.next = left & right
            Bitor.next = left | right
            Bitxor.next = left ^ right
            FloorDiv.next = 0
            if right != 0:
                FloorDiv.next = left // right
    # #         if left < 256 and right < 40:
            LeftShift.next = 0
            if left < 256 and right < 26: # fails in ghdl for > 26
                LeftShift.next = left << right
            Modulo.next = 0
            if right != 0:
                Modulo.next = left % right
            Mul.next = left * right
            # Icarus doesn't support ** yet
            # if left < 256 and right < 40:
            #    Pow.next = left ** right
    # #         Pow.next = 0
            RightShift.next = left >> right
            Sub.next = 0
            if left >= right:
                Sub.next = left - right
            Sum.next = left + right
            EQ.next = left == right
            NE.next = left != right
            LT.next = left < right
            GT.next = left > right
            LE.next = left <= right
            GE.next = left >= right
            Booland.next = bool(left) and bool(right)
            Boolor.next = bool(left) or bool(right)

    return comb


@block
def binaryBench(m, n):

    M = 2 ** m
    N = 2 ** n
    P = min(M, N)
    seqP = tuple(range(P))
    seqM = tuple([randrange(M) for __ in range(NRTESTS)])
    seqN = tuple([randrange(N) for __ in range(NRTESTS)])

    left = Signal(intbv(0)[m:])
    right = Signal(intbv(0)[n:])
    Bitand = Signal(intbv(0)[max(m, n):])
    Bitor = Signal(intbv(0)[max(m, n):])
    Bitxor = Signal(intbv(0)[max(m, n):])
    FloorDiv = Signal(intbv(0)[m:])
    LeftShift = Signal(intbv(0)[64:])
    Modulo = Signal(intbv(0)[m:])
    Mul = Signal(intbv(0)[m + n:])
    Pow = Signal(intbv(0)[64:])
    RightShift = Signal(intbv(0)[m:])
    Sub = Signal(intbv(0)[max(m, n):])
    Sum = Signal(intbv(0)[max(m, n) + 1:])
    EQ, NE, LT, GT, LE, GE = [Signal(bool()) for __ in range(6)]
    Booland, Boolor = [Signal(bool()) for i in range(2)]

    binops = binaryOps(Bitand,
                       Bitor,
                       Bitxor,
                       FloorDiv,
                       LeftShift,
                       Modulo,
                       Mul,
                       Pow,
                       RightShift,
                       Sub,
                       Sum,
                       EQ,
                       NE,
                       LT,
                       GT,
                       LE,
                       GE,
                       Booland,
                       Boolor,
                       left, right)

    @instance
    def stimulus():
        left.next = 1
        right.next = 1
        yield delay(10)
        left.next = 0
        right.next = 0
        yield delay(10)
        left.next = 0
        right.next = N - 1
        yield delay(10)
        left.next = M - 1
        right.next = 0
        yield delay(10)
        left.next = M - 1
        right.next = N - 1
        for i in range(len(seqP)):
            left.next = seqP[i]
            right.next = seqP[i]
            yield delay(10)
        for i in range(NRTESTS):
            left.next = seqM[i]
            right.next = seqN[i]
            yield delay(10)
        # raise StopSimulation

    @instance
    def check():
        while True:
            yield left, right
            yield delay(1)

            print(Bitand)
            print(Bitor)
            print(Bitxor)
            print(FloorDiv)
            print(LeftShift)

            # print Pow, Pow_v

            print(Modulo)
            print(RightShift)
            print(Mul)
            print(Sub)
            print(Sum)
            print(int(EQ))
            print(int(NE))
            print(int(LT))
            print(int(GT))
            print(int(LE))
            print(int(GE))
            print(int(Booland))
            print(int(Boolor))

    return binops, stimulus, check


def testBinary():
    for m, n in ((4, 4,), (5, 3), (2, 6), (8, 7)):
    # for m, n in ((2, 6),):
        assert binaryBench(m, n).verify_convert() == 0


@block
def multiOps(
              Bitand,
              Bitor,
              Bitxor,
              Booland,
              Boolor,
              argm, argn, argp):

    @instance
    def comb():
        while 1:
            yield argm, argn, argp
            Bitand.next = argm & argn & argp
            Bitor.next = argm | argn | argp
            Bitxor.next = argm ^ argn ^ argp
            Booland.next = bool(argm) and bool(argn) and bool(argp)
            Boolor.next = bool(argm) and bool(argn) and bool(argp)

    return comb


@block
def multiBench(m, n, p):

    M = 2 ** m
    N = 2 ** n
    P = 2 ** p

    Q = min(M, N, P)
    seqQ = tuple(range(1, Q))
    seqM = tuple([randrange(M) for __ in range(NRTESTS)])
    seqN = tuple([randrange(N) for __ in range(NRTESTS)])
    seqP = tuple([randrange(P) for __ in range(NRTESTS)])

    argm = Signal(intbv(0)[m:])
    argn = Signal(intbv(0)[n:])
    argp = Signal(intbv(0)[p:])
    Bitand = Signal(intbv(0)[max(m, n, p):])
    Bitor = Signal(intbv(0)[max(m, n, p):])
    Bitxor = Signal(intbv(0)[max(m, n, p):])
    Booland, Boolor = [Signal(bool()) for i in range(2)]

    multiops = multiOps(Bitand,
                        Bitor,
                        Bitxor,
                        Booland,
                        Boolor,
                        argm, argn, argp)

    @instance
    def stimulus():
        for i in range(len(seqQ)):
            argm.next = seqQ[i]
            argn.next = seqQ[i]
            argp.next = seqQ[i]
            yield delay(10)
        for i in range(NRTESTS):
            argm.next = seqM[i]
            argn.next = seqN[i]
            argp.next = seqP[i]
            yield delay(10)
# #         for j, k, l in ((0, 0, 0),   (0, 0, P-1), (0, N-1, P-1),
# #                         (M-1, 0, 0),  (M-1, 0, P-1), (M-1, N-1, 0),
# #                         (0, N-1, 0), (M-1, N-1, P-1)):
# #             argm.next = j
# #             argn.next = k
# #             argp.next = l
# #             yield delay(10)

    @instance
    def check():
        while 1:
            yield argm, argn, argp
            yield delay(1)

            print(Bitand)
            print(Bitor)
            print(Bitxor)
            print(int(Booland))
            print(int(Boolor))

    return multiops, stimulus, check


def testMultiOps():
    for m, n, p in ((4, 4, 4,), (5, 3, 2), (3, 4, 6), (3, 7, 4)):
        assert multiBench(m, n, p).verify_convert() == 0


@block
def unaryOps(
             Not_kw,
             Invert,
             UnaryAdd,
             UnarySub,
             arg):

    @instance
    def comb():
        while 1:
            yield arg
            Not_kw.next = not arg
            Invert.next = ~arg

            # unary operators not supported ?
            # UnaryAdd.next = +arg
            # UnarySub.next = --arg
    return comb


@block
def unaryBench(m):

    M = 2 ** m
    seqM = tuple([randrange(M) for i in range(NRTESTS)])

    arg = Signal(intbv(0)[m:])
    Not_kw = Signal(bool(0))
    Invert = Signal(intbv(0)[m:])
    UnaryAdd = Signal(intbv(0)[m:])
    UnarySub = Signal(intbv(0)[m:])

    unaryops = unaryOps(Not_kw,
                        Invert,
                        UnaryAdd,
                        UnarySub,
                        arg)

    @instance
    def stimulus():
        for i in range(NRTESTS):
            arg.next = seqM[i]
            yield delay(10)
        raise StopSimulation

    @instance
    def check():
        while 1:
            yield arg
            yield delay(1)
            print(int(Not_kw))
            print(Invert)
            # check unary operator support in vhdl
            # print UnaryAdd
            # print UnarySub

    return unaryops, stimulus, check


def testUnaryOps():
    for m in (4, 7):
        assert unaryBench(m).verify_convert() == 0


@block
def augmOps(Bitand,
              Bitor,
              Bitxor,
              FloorDiv,
              LeftShift,
              Modulo,
              Mul,
              RightShift,
              Sub,
              Sum,
              left, right):

    @instance
    def comb():
        # var = intbv(0)[min(64, len(left) + len(right)):]
        var = intbv(0)[len(left) + len(right):]
        var2 = intbv(0)[64:]
        while True:
            yield left, right
            var[:] = left
            var &= right
            Bitand.next = var
            var[:] = left
            var |= right
            Bitor.next = var
            var[:] = left
            var ^= left
            Bitxor.next = var
            FloorDiv.next = 0
            if right != 0:
                var[:] = left
                var //= right
                FloorDiv.next = var
            Sub.next = 0
            if left >= right:
                var[:] = left
                var -= right
                Sub.next = var
            var[:] = left
            var += right
            Sum.next = var
            LeftShift.next = 0
            if left < 256 and right < 26:
                var2[:] = left
                var2 <<= right
                LeftShift.next = var2
            Modulo.next = 0
            if right != 0:
                var[:] = left
                var %= right
                Modulo.next = var
            var[:] = left
            var *= right
            Mul.next = var
            var[:] = left
            var >>= right
            RightShift.next = var

    return comb


@block
def augmBench(m, n):

    M = 2 ** m
    N = 2 ** n

    seqM = tuple([randrange(M) for __ in range(NRTESTS)])
    seqN = tuple([randrange(N) for __ in range(NRTESTS)])

    left = Signal(intbv(0)[m:])
    right = Signal(intbv(0)[n:])
    Bitand = Signal(intbv(0)[max(m, n):])
    Bitor = Signal(intbv(0)[max(m, n):])
    Bitxor = Signal(intbv(0)[max(m, n):])
    FloorDiv = Signal(intbv(0)[m:])
    LeftShift = Signal(intbv(0)[64:])
    Modulo = Signal(intbv(0)[m:])
    Mul = Signal(intbv(0)[m + n:])
    RightShift = Signal(intbv(0)[m:])
    Sub = Signal(intbv(0)[max(m, n):])
    Sum = Signal(intbv(0)[max(m, n) + 1:])

    augmops = augmOps(Bitand,
                       Bitor,
                       Bitxor,
                       FloorDiv,
                       LeftShift,
                       Modulo,
                       Mul,
                       RightShift,
                       Sub,
                       Sum,
                       left, right)

    @instance
    def stimulus():
        left.next = 1
        right.next = 1
        yield delay(10)
        left.next = 0
        right.next = 0
        yield delay(10)
        left.next = 0
        right.next = N - 1
        yield delay(10)
        left.next = M - 1
        right.next = 0
        yield delay(10)
        left.next = M - 1
        right.next = N - 1
        for i in range(NRTESTS):
            left.next = seqM[i]
            right.next = seqN[i]
            yield delay(10)

    @instance
    def check():
        while True:
            yield left, right
            yield delay(1)
            print(Bitand)
            print(Bitor)
            print(Bitxor)
            print(Sub)
            print(Sum)
            print(FloorDiv)
            print(LeftShift)
            print(Modulo)
            print(Mul)
            print(RightShift)

    return augmops, stimulus, check


def testAugmOps():
    for m, n in ((4, 4,), (5, 3), (2, 6), (8, 7)):
        assert augmBench(m, n).verify_convert() == 0




#  This file is part of the myhdl library, a Python package for using
#  Python as a Hardware Description Language.
#
#  Copyright (C) 2003-2008 Jan Decaluwe
#
#  The myhdl library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public License as
#  published by the Free Software Foundation; either version 2.1 of the
#  License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful, but
#  WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.

#  You should have received a copy of the GNU Lesser General Public
#  License along with this library; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA

""" Module with the always function. """
import inspect
from types import FunctionType

from myhdl import InstanceError
from myhdl._util import _isGenFunc, _makeAST
from myhdl._Waiter import _inferWaiter
from myhdl._resolverefs import _AttrRefTransformer
from myhdl._visitors import _SigNameVisitor


class _error:
    pass
_error.NrOfArgs = "decorated generator function should not have arguments"
_error.ArgType = "decorated object should be a generator function"


class _CallInfo(object):

    def __init__(self, name, modctxt, symdict):
        self.name = name
        self.modctxt = modctxt
        self.symdict = symdict


def _getCallInfo():
    """Get info on the caller of an Instantiator.

    An Instantiator should be used in a block context.
    This function gets the required info about the caller.
    It uses the frame stack:
    0: this function
    1: the instantiator decorator
    2: the block function that defines instances
    3: the caller of the block function, e.g. the BlockInstance.
    """
    from myhdl import _block
    funcrec = inspect.stack()[2]
    name = funcrec[3]
    frame = funcrec[0]
    symdict = dict(frame.f_globals)
    symdict.update(frame.f_locals)
    modctxt = False
    callerrec = inspect.stack()[3]
    f_locals = callerrec[0].f_locals
    if 'self' in f_locals:
        modctxt = isinstance(f_locals['self'], _block._Block)
    return _CallInfo(name, modctxt, symdict)


def instance(genfunc):
    callinfo = _getCallInfo()
    if not isinstance(genfunc, FunctionType):
        raise InstanceError(_error.ArgType)
    if not _isGenFunc(genfunc):
        raise InstanceError(_error.ArgType)
    if genfunc.__code__.co_argcount > 0:
        raise InstanceError(_error.NrOfArgs)
    return _Instantiator(genfunc, callinfo=callinfo)


class _Instantiator(object):

    def __init__(self, genfunc, callinfo):
        self.callinfo = callinfo
        self.callername = callinfo.name
        self.modctxt = callinfo.modctxt
        self.genfunc = genfunc
        self.gen = genfunc()
        # infer symdict
        f = self.funcobj
        varnames = f.__code__.co_varnames
        symdict = {}
        for n, v in callinfo.symdict.items():
            if n not in varnames:
                symdict[n] = v
        self.symdict = symdict

        # print modname, genfunc.__name__
        tree = self.ast
        # print ast.dump(tree)
        v = _AttrRefTransformer(self)
        v.visit(tree)
        v = _SigNameVisitor(self.symdict)
        v.visit(tree)
        self.inputs = v.inputs
        self.outputs = v.outputs
        self.inouts = v.inouts
        self.embedded_func = v.embedded_func
        self.sigdict = v.sigdict
        self.losdict = v.losdict

    @property
    def name(self):
        return self.funcobj.__name__

    @property
    def funcobj(self):
        return self.genfunc

    @property
    def waiter(self):
        return self._waiter()(self.gen)

    def _waiter(self):
        return _inferWaiter

    @property
    def ast(self):
        return _makeAST(self.funcobj)



import os
import unittest

from myhdl import (block, Signal, intbv, instance)
from myhdl import ConversionError
from myhdl.conversion._misc import _error


class TestNotSupported(unittest.TestCase):

    def check(self, *args):
        try:
            args[0](*args[1:]).convert(hdl='Verilog')
        except ConversionError as e:
            self.assertEqual(e.kind, _error.NotSupported)
        except Exception as e:
            raise e

    def nocheck(self, *args):
        args[0](*args[1:]).convert(hdl='Verilog')

    def testAssList(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    [p, q] = 1, 2

            return comb

        self.check(g, z, a)

    def testAssTuple(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    p, q = 1, 2

            return comb

        self.check(g, z, a)

    def testClass(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1

                    class c:
                        pass

            return comb

        self.check(g, z, a)

    def testDict(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    d = {}

            return comb

        self.check(g, z, a)

    def testDiv(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = z / a

            return comb

        self.check(g, z, a)

    def testFrom(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    from os import path

            return comb

        self.check(g, z, a)

    def testFunction(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1

                    def f():
                        pass

            return comb

        self.check(g, z, a)

    def testGlobal(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    global e

            return comb

        self.check(g, z, a)

    def testImport(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    import os

            return comb

        self.check(g, z, a)

    def testLambda(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    lambda: 1

            return comb

        self.check(g, z, a)

    def testListCompIf(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    l = [i for i in range(5) if i > 1]

            return comb

        self.check(g, z, a)

    def testList(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    l = [1, 2, 3]

            return comb

        self.check(g, z, a)

# #     def testPower(self):
# #         a = Signal(bool())
# #         z = Signal(bool())
# #         def g(z, a):
# #             while 1:
# #                 yield a
# #                 z.next = 2 ** 8
# #         self.check(g, z, a)

# #     def testReturn(self):
# #         a = Signal(bool())
# #         z = Signal(bool())
# #         def g(z, a):
# #             while 1:
# #                 yield a
# #                 z.next = 1
# #                 return
# #         self.check(g, z, a)

    def testTryExcept(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    try:
                        pass
                    except:
                        pass

            return comb

        self.check(g, z, a)

    def testTryFinally(self):
        a = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a):

            @instance
            def comb():
                while 1:
                    yield a
                    z.next = 1
                    try:
                        pass
                    finally:
                        pass

            return comb

        self.check(g, z, a)

    def testChainedCompare(self):
        a, b, c = [Signal(bool()) for i in range(3)]
        z = Signal(bool())

        @block
        def g(z, a, b, c):

            @instance
            def comb():
                while 1:
                    yield a, b, c
                    z.next = a <= b < c

            return comb

        self.check(g, z, a, b, c)

# #     def testShortcutAnd(self):
# #         a, b = [Signal(bool()) for i in range(2)]
# #         z = Signal(bool())
# #         def g(z, a, b):
# #             while 1:
# #                 yield a
# #                 if a:
# #                     pass
# #                 else:
# #                     z.next = a and b
# #         self.check(g, z, a, b)

# #     def testShortcutOr(self):
# #         a, b, c  = [Signal(bool()) for i in range(3)]
# #         z = Signal(bool())
# #         def g(z, a, b):
# #             while 1:
# #                 yield a
# #                 if a:
# #                     pass
# #                 else:
# #                     z.next = a < (b or c)
# #         self.check(g, z, a, b)

    def testNonBoolArgAnd(self):
        a = Signal(bool())
        b = intbv(0)[2:]
        z = Signal(bool())

        @block
        def g(z, a, b):

            @instance
            def comb():
                while 1:
                    yield a
                    if a:
                        pass
                    else:
                        z.next = a and b

            return comb

        self.check(g, z, a, b)

    def testNonBoolArgOr(self):
        a = Signal(bool())
        b = intbv(0)[2:]
        c = Signal(bool())
        z = Signal(bool())

        @block
        def g(z, a, b):

            @instance
            def comb():
                while 1:
                    yield a
                    if a:
                        pass
                    else:
                        z.next = a < (b or c)

            return comb

        self.check(g, z, a, b)

    def testExtraArguments(self):
        a, b, c = [Signal(bool()) for i in range(3)]
        c = [1, 2]

        def g(a, *args):
            return a

        @block
        def f(a, b, c, *args):

            @instance
            def comb():
                g(a, b)
                yield a

            return comb

        self.check(f, a, b, c)

    def testExtraPositionalArgsInCall(self):
        a, b, c = [Signal(bool()) for __ in range(3)]
        c = [1]
        d = {'b':2}

        def h(b):
            return b

        @block
        def g(a):

            @instance
            def comb():
                h(*c)
                yield a

            return comb

        @block
        def f(a, b, c):
            return g(a)

        x = self.check(f, a, b, c)

    def testExtraNamedArgsInCall(self):
        a, b, c = [Signal(bool()) for __ in range(3)]
        c = [1]
        d = {'b':2}

        def h(b):
            return b

        @block
        def g(a):

            @instance
            def comb():
                h(**d)
                yield a

            return comb

        @block
        def f(a, b, c):
            return g(a)

        x = self.check(f, a, b, c)


class TestMisc(unittest.TestCase):

    def test(self):
        a, b, c = [Signal(bool()) for __ in range(3)]
        c = [1]
        d = {'a':2}

        def h(b):
            return b

        @block
        def g(a):

            @instance
            def comb():
                h(a)
                yield a

            return comb

        # b and c are not being used
        @block
        def f(a, b, c):
            return g(a)

        f(a, b, c)
        x = f(a, b, c).convert(hdl='Verilog')

